import asyncio
import sys
import time
import uuid
import warnings
from concurrent.futures import CancelledError
from dataclasses import dataclass
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    Literal,
    Optional,
    Union,
    overload,
)

from fused._options import options as OPTIONS

from . import load_async

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd
    import xarray as xr

from loguru import logger

from fused._load_udf import load
from fused._udf.decorators import _parse_ttl
from fused.models.api import UdfAccessToken, UdfJobStepConfig
from fused.models.api.udf_access_token import is_udf_token
from fused.models.request import (
    WHITELISTED_INSTANCE_TYPES,
    WHITELISTED_INSTANCE_TYPES_values,
)
from fused.models.udf import Udf
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.types import UdfRuntimeError, UdfSerializationError, UdfTimeoutError
from fused.warnings import FusedDeprecationWarning, FusedIgnoredWarning, FusedWarning

from .core import (
    run_shared_file,
    run_shared_file_async,
    run_shared_tile,
    run_shared_tile_async,
)
from .core._impl._realtime_ops_impl import default_run_engine

ResultType = Union["xr.Dataset", "pd.DataFrame", "gpd.GeoDataFrame"]

InstanceType = Literal[
    "realtime", "small", "medium", "large", WHITELISTED_INSTANCE_TYPES
]


RUN_KWARGS = {
    "x",
    "y",
    "z",
    "sync",
    "engine",
    "type",
    "max_retry",
    "cache_max_age",
    "cache",
    "verbose",
    "parameters",
    "_format",
    "_return_response",
    "_profile",
    "_use_polling",
}


@dataclass
class ResolvedTokenBasedUdf:
    """Represents a UDF resolved as a token for remote execution."""

    token: str
    storage_type: Literal["token"] = "token"
    udf: str = ""  # token string

    def __post_init__(self):
        if not self.udf:
            self.udf = self.token


@dataclass
class ResolvedLocalJobStepUdf:
    """Represents a UDF resolved as a local job step."""

    job_step: UdfJobStepConfig
    udf: Udf
    storage_type: Literal["local_job_step"] = "local_job_step"


ResolvedUdf = Union[ResolvedTokenBasedUdf, ResolvedLocalJobStepUdf]


async def resolve_udf_async(
    udf: Union[str, None, UdfJobStepConfig, Udf, UdfAccessToken],
    engine: str,
    is_token: bool,
) -> ResolvedUdf:
    """
    Async version of resolve_udf that handles I/O operations without blocking.

    Resolve UDF input into standardized components, handling all loading scenarios.

    Args:
        udf: The UDF to resolve
        engine: Execution engine (used for local loading decisions)
        is_token: Whether the UDF is a token

    Returns:
        UdfResolution object containing resolved UDF information
    """

    if udf is None:
        raise ValueError("No UDF specified")

    elif is_token:
        # Handle local engine loading for tokens
        token = udf if isinstance(udf, str) else udf.token
        if engine == "local":
            job_step, loaded_udf = await _load_udf_and_setup_job_step_async(token)
            return ResolvedLocalJobStepUdf(job_step=job_step, udf=loaded_udf)
        else:
            return ResolvedTokenBasedUdf(token=token)

    elif isinstance(udf, Udf):
        job_step = UdfJobStepConfig(udf=udf)
        return ResolvedLocalJobStepUdf(job_step=job_step, udf=udf)

    elif isinstance(udf, str):
        job_step, loaded_udf = await _load_udf_and_setup_job_step_async(udf)
        return ResolvedLocalJobStepUdf(job_step=job_step, udf=loaded_udf)

    else:
        raise TypeError(
            "Could not detect UDF from first parameter. It should be a string or Udf object."
        )


def resolve_udf(
    udf: Union[str, None, UdfJobStepConfig, Udf, UdfAccessToken],
    engine: str,
    is_token: bool,
) -> ResolvedUdf:
    """
    Resolve UDF input into standardized components, handling all loading scenarios.

    Args:
        udf: The UDF to resolve
        engine: Execution engine (used for local loading decisions)
        is_token: Whether the UDF is a token

    Returns:
        UdfResolution object containing resolved UDF information
    """
    if udf is None:
        raise ValueError("No UDF specified")

    elif is_token:
        # Handle local engine loading for tokens
        token = udf if isinstance(udf, str) else udf.token
        if engine == "local":
            job_step, loaded_udf = _load_udf_and_setup_job_step(token)
            return ResolvedLocalJobStepUdf(job_step=job_step, udf=loaded_udf)
        else:
            return ResolvedTokenBasedUdf(token=token)

    elif isinstance(udf, Udf):
        job_step = UdfJobStepConfig(udf=udf)
        return ResolvedLocalJobStepUdf(job_step=job_step, udf=udf)

    elif isinstance(udf, str):
        job_step, loaded_udf = _load_udf_and_setup_job_step(udf)
        return ResolvedLocalJobStepUdf(job_step=job_step, udf=loaded_udf)

    else:
        raise TypeError(
            "Could not detect UDF from first parameter. It should be a string or Udf object."
        )


def _load_udf_and_setup_job_step(udf_source: str) -> tuple[UdfJobStepConfig, Udf]:
    """Helper to load UDF and create job step configuration."""
    try:
        loaded_udf = load(udf_source)
    except Exception as exc:
        raise ValueError(
            "Could not load UDF. Make sure the UDF is available locally when "
            'using `engine="local"`.\n'
            f'Error loading the UDF: "{exc}"'
        )

    job_step = UdfJobStepConfig(udf=loaded_udf)
    return job_step, loaded_udf


async def _load_udf_and_setup_job_step_async(
    udf_source: str,
) -> tuple[UdfJobStepConfig, Udf]:
    """Async helper to load UDF and create job step configuration."""

    try:
        loaded_udf = await load_async(udf_source)
    except Exception as exc:
        raise ValueError(
            "Could not load UDF. Make sure the UDF is available locally when "
            'using `engine="local"`.\n'
            f'Error loading the UDF: "{exc}"'
        )

    job_step = UdfJobStepConfig(udf=loaded_udf)
    return job_step, loaded_udf


def _create_local_params(
    local_tile_bbox: Optional["gpd.GeoDataFrame"],
    cache_max_age: Optional[int],
    _return_response: Optional[bool],
    parameters: Dict[str, Any],
    include_bbox: bool = True,
) -> Dict[str, Any]:
    """Helper to create common local execution parameters."""
    params = {
        "cache_max_age": cache_max_age,
        "_return_response": _return_response,
        **parameters,
    }
    if include_bbox:
        # For tile operations, include the bbox as first positional arg
        return {"bbox": local_tile_bbox, **params}
    return params


def _resolve_engine(
    engine: Optional[Literal["remote", "local"]],
    is_token: bool,
    instance_type: Optional[str],
) -> Literal["remote", "local"]:
    """Determine the execution engine based on parameters and storage type."""
    if engine == "realtime":
        engine = "remote"
    elif engine is None:
        if is_token:
            engine = "remote"
        else:
            engine = default_run_engine()
    elif engine not in ("local", "remote"):
        raise ValueError("Invalid engine specified. Must be 'local' or 'remote'.")

    if instance_type is not None:
        if engine == "local":
            raise ValueError(
                "Specifying an instance type is only supported for the 'remote' engine."
            )
        if (
            instance_type not in ("realtime", "small", "medium", "large")
            and instance_type not in WHITELISTED_INSTANCE_TYPES_values
        ):
            raise ValueError(
                "Invalid instance type specified. Must be 'realtime', 'small', "
                "'medium', 'large', or one of the whitelisted instance types "
                f"({WHITELISTED_INSTANCE_TYPES_values})."
            )

    return engine


def _resolve_instance_type(
    instance_type: Optional[str], resolved_udf: ResolvedUdf
) -> str:
    # if not specified in run(), use the default defined on the UDF
    if instance_type is None and resolved_udf.storage_type == "local_job_step":
        instance_type = resolved_udf.udf.instance_type

    if instance_type is None:
        # TODO potentially allow to override the default instance type with an option?
        instance_type = "realtime"

    if resolved_udf.storage_type == "token" and instance_type != "realtime":
        # TODO we could in theory allow to run shared tokens that can be loaded
        # locally, or support this on the server side
        raise ValueError(
            "UDF shared tokens can only be run with the 'realtime' instance type"
        )

    return instance_type


def _process_tile_coordinates(
    x: Optional[int],
    y: Optional[int],
    z: Optional[int],
    type: Optional[Literal["tile", "file"]],
    kw_parameters: Dict[str, Any],
) -> tuple[Optional["gpd.GeoDataFrame"], bool, Optional[Literal["tile", "file"]]]:
    """Process tile coordinates and determine execution type."""
    from fused._optional_deps import HAS_GEOPANDAS, HAS_MERCANTILE, HAS_SHAPELY

    local_tile_bbox: Optional["gpd.GeoDataFrame"] = None
    xyz_ignored = False

    if x is not None and y is not None and z is not None:
        if HAS_MERCANTILE and HAS_GEOPANDAS and HAS_SHAPELY:
            import geopandas as gpd
            import mercantile
            import shapely

            tile_bounds = mercantile.bounds(x, y, z)
            local_tile_bbox = gpd.GeoDataFrame(
                {"x": [x], "y": [y], "z": [z]},
                geometry=[shapely.box(*tile_bounds)],
                crs=4326,
            )
        else:
            xyz_ignored = True

    if x is not None and y is not None and z is not None:
        if type != "tile":
            if type is None:
                # by default we still interpret x/y/z (if all passed) as a tile run
                type = "tile"
            else:
                kw_parameters["x"] = x
                kw_parameters["y"] = y
                kw_parameters["z"] = z
    else:
        if type is None:
            type = "file"
        elif type != "file":
            raise ValueError(
                "x, y, z not specified but type is 'tile', which is an invalid configuration. You must specify x, y, and z."
            )
        if x is not None:
            kw_parameters["x"] = x
        if y is not None:
            kw_parameters["y"] = y
        if z is not None:
            kw_parameters["z"] = z

    return local_tile_bbox, xyz_ignored, type


def _process_parameters(
    parameters: Optional[Dict[str, Any]],
    kw_parameters: Dict[str, Any],
    udf: Union[str, Udf],
    type: Optional[Literal["tile", "file"]],
    _ignore_unknown_arguments: bool,
) -> tuple[Dict[str, Any], bool]:
    """Process and validate parameters."""
    verbose = kw_parameters.get("verbose", None)
    if verbose is None:
        verbose = OPTIONS.verbose_udf_runs

    # TMP handle backwards compatibility for specifying (_)dtype_out_raster/vector
    dtype_out_raster = kw_parameters.pop(
        "dtype_out_raster", kw_parameters.pop("_dtype_out_raster", None)
    )
    dtype_out_vector = kw_parameters.pop(
        "dtype_out_vector", kw_parameters.pop("_dtype_out_vector", None)
    )
    format = kw_parameters.pop("_format", None)
    if dtype_out_raster is not None or dtype_out_vector is not None:
        warnings.warn(
            "The `dtype_out_raster` and `dtype_out_vector` parameters are deprecated "
            "and will be removed in a future release.",
            FusedDeprecationWarning,
            stacklevel=3,
        )
        if format is None:
            if dtype_out_raster is not None and dtype_out_vector is not None:
                format = f"{dtype_out_raster},{dtype_out_vector}"
            elif dtype_out_raster is not None:
                format = dtype_out_raster
            elif dtype_out_vector is not None:
                format = dtype_out_vector

    if format is not None:
        kw_parameters["_format"] = format

    merged_parameters = {
        **kw_parameters,
        **(parameters if parameters is not None else {}),
    }

    # Raise an error if passing a different argument than those expected from
    # the union of the UDF own parameters and the run() function parameters
    if merged_parameters and not _ignore_unknown_arguments:
        if isinstance(udf, Udf) and udf._parameter_list:
            udf_kwargs = set(udf._parameter_list)
            allowed_kwargs = udf_kwargs.union(
                RUN_KWARGS if type == "tile" else RUN_KWARGS - {"x", "y", "z"}
            )
            kw_params = set(merged_parameters.keys())

            if not kw_params.issubset(allowed_kwargs) and not udf._parameter_has_kwargs:
                unexpected_args = kw_params - allowed_kwargs
                raise TypeError(
                    f"{udf.entrypoint}() got unexpected keyword argument '{unexpected_args.pop()}'"
                )

    return merged_parameters, verbose


def _create_execution_function(
    dispatch_params: tuple,
    udf_resolution: ResolvedUdf,
    x: Optional[int],
    y: Optional[int],
    z: Optional[int],
    cache_max_age: Optional[int],
    local_tile_bbox: Optional["gpd.GeoDataFrame"],
    _return_response: Optional[bool],
    parameters: Dict[str, Any],
    run_id: str,
) -> Callable:
    """Create the appropriate execution function based on dispatch parameters."""
    exec_type, storage_type, engine, instance_type = dispatch_params

    # Prepare parameter sets
    tile_params = {
        "x": x,
        "y": y,
        "z": z,
        "cache_max_age": cache_max_age,
        "_run_id": run_id,
        **parameters,
    }
    file_params = {
        "cache_max_age": cache_max_age,
        "_run_id": run_id,
        **parameters,
    }

    # Shared UDF token
    if storage_type == "token":
        assert engine == "remote"
        assert isinstance(udf_resolution, ResolvedTokenBasedUdf)
        if exec_type == "tile":
            return partial(run_shared_tile, udf_resolution.token, **tile_params)
        else:  # file
            return partial(run_shared_file, udf_resolution.token, **file_params)

    # Local job step
    elif storage_type == "local_job_step":
        assert isinstance(udf_resolution, ResolvedLocalJobStepUdf)
        job_step = udf_resolution.job_step
        if engine == "remote" and instance_type == "realtime":
            if exec_type == "tile":
                return lambda: job_step._run_tile(**tile_params)
            else:  # file
                # Async file doesn't use cache_max_age for some reason
                return lambda: job_step._run_file(**file_params)
        elif engine == "remote":  # one of the batch instance types
            return lambda: job_step.set_udf(
                job_step.udf, parameters=parameters
            )._run_batch(instance_type=instance_type)
        else:  # local engine
            if exec_type == "tile":
                local_params = _create_local_params(
                    local_tile_bbox,
                    cache_max_age,
                    _return_response,
                    parameters,
                    include_bbox=True,
                )
                return lambda: job_step._run_local(
                    local_tile_bbox,
                    **{k: v for k, v in local_params.items() if k != "bbox"},
                )
            else:  # file
                local_params = _create_local_params(
                    local_tile_bbox,
                    cache_max_age,
                    _return_response,
                    parameters,
                    include_bbox=False,
                )
                return lambda: job_step._run_local(**local_params)

    else:
        # in theory we should never get here
        raise ValueError(f"Call type is not yet implemented: {dispatch_params}")


def _create_execution_function_async(
    dispatch_params: tuple,
    udf_resolution: ResolvedUdf,
    x: Optional[int],
    y: Optional[int],
    z: Optional[int],
    cache_max_age: Optional[int],
    local_tile_bbox: Optional["gpd.GeoDataFrame"],
    _return_response: Optional[bool],
    parameters: Dict[str, Any],
    run_id: str,
) -> Callable:
    """Create the appropriate async execution function based on dispatch parameters."""
    exec_type, storage_type, engine, instance_type = dispatch_params

    # Prepare parameter sets
    tile_params = {
        "x": x,
        "y": y,
        "z": z,
        "cache_max_age": cache_max_age,
        "_run_id": run_id,
        **parameters,
    }
    file_params = {
        "cache_max_age": cache_max_age,
        "_run_id": run_id,
        **parameters,
    }

    # Shared UDF token
    if storage_type == "token":
        assert engine == "remote"
        assert isinstance(udf_resolution, ResolvedTokenBasedUdf)
        if exec_type == "tile":
            return partial(run_shared_tile_async, udf_resolution.token, **tile_params)
        else:  # file
            return partial(run_shared_file_async, udf_resolution.token, **file_params)

    # Local job step
    elif storage_type == "local_job_step":
        assert isinstance(udf_resolution, ResolvedLocalJobStepUdf)
        job_step = udf_resolution.job_step
        if engine == "remote" and instance_type == "realtime":
            if exec_type == "tile":
                return lambda: job_step._run_tile_async(**tile_params)
            else:  # file
                return lambda: job_step._run_file_async(**file_params)
        elif engine == "remote":  # one of the batch instance types

            async def _async_batch():
                return job_step.set_udf(job_step.udf, parameters=parameters)._run_batch(
                    instance_type=instance_type
                )

            return _async_batch
        else:  # local engine
            # Local execution is inherently sync, so we wrap it in an async function
            if exec_type == "tile":
                local_params = _create_local_params(
                    local_tile_bbox,
                    cache_max_age,
                    _return_response,
                    parameters,
                    include_bbox=True,
                )

                async def _async_tile_local():
                    return job_step._run_local(
                        local_tile_bbox,
                        **{k: v for k, v in local_params.items() if k != "bbox"},
                    )

                return _async_tile_local
            else:  # file
                local_params = _create_local_params(
                    local_tile_bbox,
                    cache_max_age,
                    _return_response,
                    parameters,
                    include_bbox=False,
                )

                async def _async_file_local():
                    return job_step._run_local(**local_params)

                return _async_file_local

    else:
        raise ValueError(f"Call type is not yet implemented: {dispatch_params}")


async def _execute_with_retry_async(
    fn: Callable[[], Awaitable[Any]],
    max_retry: int,
    _cancel_callback: Callable[[], bool] | None,
    _return_response: bool,
    verbose: bool,
) -> Any:
    """Run `fn` with exponential backoff.

    Cancels immediately if `_cancel_callback` returns True. Returns the
    `_process_result` of `fn()`; raises the last error after retries.
    """

    n_retries = 0
    while n_retries <= max_retry:
        cancel_requested = _cancel_callback is not None and _cancel_callback()
        try:
            if cancel_requested:
                raise CancelledError("Cancel requested")

            udf_eval_result = await fn()
            # Nested and remote UDF calls will return UdfEvaluationResult.
            # merge the stdout/stderr from fused.run() into running environment,
            # then return the UdfEvaluationResult object.
            return _process_result(udf_eval_result, _return_response, verbose)
        except Exception as exc:
            if (
                isinstance(
                    exc, (UdfSerializationError, UdfRuntimeError, CancelledError)
                )
                or n_retries >= max_retry
            ):
                raise

            delay = OPTIONS.request_retry_base_delay * (2**n_retries)
            n_retries += 1
            warnings.warn(
                f"UDF execution failed, retrying in {delay} seconds (error: {exc})",
                FusedWarning,
            )

            await asyncio.sleep(delay)


def _execute_with_retry(
    fn: Callable[[], Any],
    max_retry: int,
    _cancel_callback: Callable[[], bool] | None,
    _return_response: bool,
    verbose: bool,
) -> Any:
    """Run `fn` with exponential backoff.

    Cancels immediately if `_cancel_callback` returns True. Returns the
    `_process_result` of `fn()`; raises the last error after retries.
    """
    n_retries = 0
    while n_retries <= max_retry:
        cancel_requested = _cancel_callback is not None and _cancel_callback()
        try:
            if cancel_requested:
                raise CancelledError("Cancel requested")
            udf_eval_result = fn()

            # Nested and remote UDF calls will return UdfEvaluationResult.
            # merge the stdout/stderr from fused.run() into running environment,
            # then return the UdfEvaluationResult object.
            return _process_result(udf_eval_result, _return_response, verbose)

        except Exception as exc:
            if (
                isinstance(
                    exc, (UdfSerializationError, UdfRuntimeError, CancelledError)
                )
                or n_retries >= max_retry
            ):
                raise

            delay = OPTIONS.request_retry_base_delay * (2**n_retries)
            n_retries += 1
            warnings.warn(
                f"UDF execution failed, retrying in {delay} seconds (error: {exc})",
                FusedWarning,
            )

            time.sleep(delay)


def _process_result(udf_eval_result, _return_response: Optional[bool], verbose: bool):
    """Process the UDF execution result and handle output."""
    # Nested and remote UDF calls will return UdfEvaluationResult.
    # merge the stdout/stderr from fused.run() into running environment,
    # then return the UdfEvaluationResult object.

    if _return_response:
        return udf_eval_result
    if isinstance(udf_eval_result, UdfEvaluationResult):
        has_error = udf_eval_result.has_exception
        if udf_eval_result.stdout and (verbose or has_error):
            sys.stdout.write(udf_eval_result.stdout)
        if udf_eval_result.stderr and (verbose or has_error):
            sys.stderr.write(udf_eval_result.stderr)
        if udf_eval_result.is_cached:
            if verbose:
                sys.stdout.write("Cached UDF result returned.\n")
            logger.info(f"Cache source: {udf_eval_result.cache_source.value}")
        if udf_eval_result.error_message is not None:
            if udf_eval_result.error_type == "serialization_error":
                raise UdfSerializationError(udf_eval_result.error_message)
            elif udf_eval_result.error_type == "timeout_error":
                raise UdfTimeoutError(udf_eval_result.error_message)
            raise UdfRuntimeError(
                udf_eval_result.error_message,
                child_exception_class=udf_eval_result.exception_class,
            )
        sys.stderr.flush()
        return udf_eval_result.data

    return udf_eval_result


@overload
def run(
    udf: Union[str, None, UdfJobStepConfig, Udf, UdfAccessToken] = None,
    *,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: Literal[True] = True,
    engine: Optional[Literal["remote", "local"]] = None,
    instance_type: Optional[
        Literal["realtime", "batch", WHITELISTED_INSTANCE_TYPES]
    ] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Literal[False] = False,
    _ignore_unknown_arguments: bool = False,
    _cancel_callback: Callable[[], bool] | None = None,
    **kw_parameters,
) -> ResultType: ...


@overload
def run(
    udf: Union[str, None, UdfJobStepConfig, Udf, UdfAccessToken] = None,
    *,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: Literal[True] = True,
    engine: Optional[Literal["remote", "local"]] = None,
    instance_type: Optional[InstanceType] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Literal[True] = True,
    _ignore_unknown_arguments: bool = False,
    _cancel_callback: Callable[[], bool] | None = None,
    **kw_parameters,
) -> UdfEvaluationResult: ...


@overload
def run(
    udf: Union[str, None, UdfJobStepConfig, Udf, UdfAccessToken] = None,
    *,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: Literal[False] = False,
    engine: Optional[Literal["remote", "local"]] = None,
    instance_type: Optional[InstanceType] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Literal[False] = False,
    _ignore_unknown_arguments: bool = False,
    _cancel_callback: Callable[[], bool] | None = None,
    **kw_parameters,
) -> Coroutine[ResultType, None, None]: ...


def run(
    udf: Union[str, None, UdfJobStepConfig, Udf, UdfAccessToken] = None,
    *,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: bool = True,
    engine: Optional[Literal["remote", "local"]] = None,
    instance_type: Optional[InstanceType] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Optional[bool] = False,
    _ignore_unknown_arguments: bool = False,
    _cancel_callback: Callable[[], bool] | None = None,
    **kw_parameters,
) -> Union[
    ResultType,
    Coroutine[ResultType, None, None],
    UdfEvaluationResult,
    Coroutine[UdfEvaluationResult, None, None],
]:
    """
    Executes a user-defined function (UDF) with various execution and input options.

    This function supports executing UDFs in different environments (local or remote),
    with different types of inputs (tile coordinates, geographical bounding boxes, etc.), and
    allows for both synchronous and asynchronous execution. It dynamically determines the execution
    path based on the provided parameters.

    Args:
        udf (str, Udf or UdfJobStepConfig): the UDF to execute.
            The UDF can be specified in several ways:
            - A string representing a UDF name or UDF shared token.
            - A UDF object.
            - A UdfJobStepConfig object for detailed execution configuration.
        x, y, z: Tile coordinates for tile-based UDF execution.
        sync: If True, execute the UDF synchronously. If False, execute asynchronously.
        engine: The execution engine to use ('remote' or 'local').
        instance_type: The type of instance to use for remote execution ('realtime',
            or 'small', 'medium', 'large' or one of the whitelisted instance types).
            If not specified, gets the default from the UDF (if specified in the
            ``@fused.udf()`` decorator, and the UDF is not run as a shared token),
            or otherwise defaults to 'realtime'.
        type: The type of UDF execution ('tile' or 'file').
        max_retry: The maximum number of retries to attempt if the UDF fails.
            By default does not retry.
        cache_max_age: The maximum age when returning a result from the cache.
            Supported units are seconds (s), minutes (m), hours (h), and days (d) (e.g. “48h”, “10s”, etc.).
            Default is `None` so a UDF run with `fused.run()` will follow `cache_max_age` defined in `@fused.udf()` unless this value is changed.
        cache: Set to False as a shortcut for `cache_max_age='0s'` to disable caching.
        verbose: Set to False to suppress any print statements from the UDF.
        parameters: Additional parameters to pass to the UDF.
        **kw_parameters: Additional parameters to pass to the UDF.

    Raises:
        ValueError: If the UDF is not specified or is specified in more than one way.
        TypeError: If the first parameter is not of an expected type.
        Warning: Various warnings are issued for ignored parameters based on the execution path chosen.

    Returns:
        The result of the UDF execution, which varies based on the UDF and execution path.

    Examples:
        Run a UDF saved in the Fused system:
        ```py
        fused.run("username@fused.io/my_udf_name")
        ```

        Run a UDF saved in GitHub:
        ```py
        loaded_udf = fused.load("https://github.com/fusedio/udfs/tree/main/public/Building_Tile_Example")
        fused.run(loaded_udf, bbox=bbox)
        ```

        Run a UDF saved in a local directory:
        ```py
        loaded_udf = fused.load("/Users/local/dir/Building_Tile_Example")
        fused.run(loaded_udf, bbox=bbox)
        ```

    Note:
        This function dynamically determines the execution path and parameters based on the inputs.
        It is designed to be flexible and support various UDF execution scenarios.
    """
    if sync is False:
        warnings.warn(
            "sync=False' parameter is deprecated, use 'run_async()' instead of 'run()'",
            FusedDeprecationWarning,
        )

        async def _r():
            return await run_async(
                udf,
                x=x,
                y=y,
                z=z,
                sync=sync,
                engine=engine,
                type=type,
                max_retry=max_retry,
                cache_max_age=cache_max_age,
                parameters=parameters,
                _return_response=_return_response,
                _ignore_unknown_arguments=_ignore_unknown_arguments,
                _cancel_callback=_cancel_callback,
                **kw_parameters,
            )

        return _r()

    run_id = str(uuid.uuid4())
    logger.debug(f"Running UDF with {run_id=}")

    # Determine execution engine first (needed for UDF resolution)
    is_token = (isinstance(udf, str) and is_udf_token(udf)) or isinstance(
        udf, UdfAccessToken
    )
    engine = _resolve_engine(engine, is_token, instance_type)

    # Resolve UDF and get all configuration
    resolved_udf = resolve_udf(udf, engine, is_token)
    instance_type = _resolve_instance_type(instance_type, resolved_udf)

    # Process tile coordinates and determine execution type
    local_tile_bbox, xyz_ignored, type = _process_tile_coordinates(
        x, y, z, type, kw_parameters
    )

    # Process and validate parameters
    parameters, verbose = _process_parameters(
        parameters, kw_parameters, resolved_udf.udf, type, _ignore_unknown_arguments
    )

    # Parsed cache TTL
    cache_max_age = _parse_ttl(cache_max_age, cache)

    # Create and execute the function
    dispatch_params = (
        type,
        resolved_udf.storage_type,
        engine,
        instance_type,
    )

    fn = _create_execution_function(
        dispatch_params,
        resolved_udf,
        x,
        y,
        z,
        cache_max_age,
        local_tile_bbox,
        _return_response,
        parameters,
        run_id,
    )

    if xyz_ignored and engine == "local":
        # This warning doesn't matter on realtime because we will just put the x/y/z into the URL
        warnings.warn(
            FusedIgnoredWarning(
                "x, y, z arguments will be ignored because the following packages were not all found: mercantile shapely geopandas"
            ),
        )

    return _execute_with_retry(
        fn, max_retry, _cancel_callback, _return_response, verbose
    )


@overload
async def run_async(
    udf: Union[str, None, UdfJobStepConfig, Udf, UdfAccessToken] = None,
    *,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    engine: Optional[Literal["remote", "local"]] = None,
    instance_type: Optional[InstanceType] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Literal[False] = False,
    _ignore_unknown_arguments: bool = False,
    _cancel_callback: Callable[[], bool] | None = None,
    **kw_parameters,
) -> ResultType: ...


@overload
async def run_async(
    udf: Union[str, None, UdfJobStepConfig, Udf, UdfAccessToken] = None,
    *,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    engine: Optional[Literal["remote", "local"]] = None,
    instance_type: Optional[InstanceType] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Literal[True] = True,
    _ignore_unknown_arguments: bool = False,
    _cancel_callback: Callable[[], bool] | None = None,
    **kw_parameters,
) -> UdfEvaluationResult: ...


async def run_async(
    udf: Union[str, None, UdfJobStepConfig, Udf, UdfAccessToken] = None,
    *,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    engine: Optional[Literal["remote", "local"]] = None,
    instance_type: Optional[InstanceType] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    parameters: Optional[Dict[str, Any]] = None,
    _return_response: Optional[bool] = False,
    _ignore_unknown_arguments: bool = False,
    _cancel_callback: Callable[[], bool] | None = None,
    **kw_parameters,
) -> Union[ResultType, UdfEvaluationResult]:
    """
    Async version of run() that executes a user-defined function (UDF) asynchronously.

    This function provides the same functionality as run() but with async execution.
    It supports executing UDFs in different environments (local or remote),
    with different types of inputs (tile coordinates, geographical bounding boxes, etc.).

    Args:
        udf (str, Udf or UdfJobStepConfig): the UDF to execute.
            The UDF can be specified in several ways:
            - A string representing a UDF name or UDF shared token.
            - A UDF object.
            - A UdfJobStepConfig object for detailed execution configuration.
        x, y, z: Tile coordinates for tile-based UDF execution.
        engine: The execution engine to use ('remote' or 'local').
        instance_type: The type of instance to use for remote execution ('realtime',
            or 'small', 'medium', 'large' or one of the whitelisted instance types).
            If not specified, gets the default from the UDF (if specified in the
            ``@fused.udf()`` decorator, and the UDF is not run as a shared token),
            or otherwise defaults to 'realtime'.
        type: The type of UDF execution ('tile' or 'file').
        max_retry: The maximum number of retries to attempt if the UDF fails.
            By default does not retry.
        cache_max_age: The maximum age when returning a result from the cache.
            Supported units are seconds (s), minutes (m), hours (h), and days (d) (e.g. "48h", "10s", etc.).
            Default is `None` so a UDF run with `fused.run_async()` will follow `cache_max_age` defined in `@fused.udf()` unless this value is changed.
        cache: Set to False as a shortcut for `cache_max_age='0s'` to disable caching.
        verbose: Set to False to suppress any print statements from the UDF.
        parameters: Additional parameters to pass to the UDF.
        **kw_parameters: Additional parameters to pass to the UDF.

    Raises:
        ValueError: If the UDF is not specified or is specified in more than one way.
        TypeError: If the first parameter is not of an expected type.
        Warning: Various warnings are issued for ignored parameters based on the execution path chosen.

    Returns:
        The result of the UDF execution, which varies based on the UDF and execution path.

    Examples:
        Run a UDF saved in the Fused system asynchronously:
        ```py
        result = await fused.run_async("username@fused.io/my_udf_name")
        ```

        Run a UDF saved in GitHub asynchronously:
        ```py
        loaded_udf = fused.load("https://github.com/fusedio/udfs/tree/main/public/Building_Tile_Example")
        result = await fused.run_async(loaded_udf, bbox=bbox)
        ```

        Run a UDF saved in a local directory asynchronously:
        ```py
        loaded_udf = fused.load("/Users/local/dir/Building_Tile_Example")
        result = await fused.run_async(loaded_udf, bbox=bbox)
        ```

    Note:
        This function always executes asynchronously. For synchronous execution, use run().
        It uses the same parameter validation and UDF resolution logic as run().
    """
    run_id = str(uuid.uuid4())

    # Determine execution engine first (needed for UDF resolution)
    is_token = (isinstance(udf, str) and is_udf_token(udf)) or isinstance(
        udf, UdfAccessToken
    )
    engine = _resolve_engine(engine, is_token, instance_type)

    # Resolve UDF and get all configuration
    resolved_udf = await resolve_udf_async(udf, engine, is_token)
    instance_type = _resolve_instance_type(instance_type, resolved_udf)

    # Process tile coordinates and determine execution type
    local_tile_bbox, xyz_ignored, type = _process_tile_coordinates(
        x, y, z, type, kw_parameters
    )

    # Process and validate parameters
    parameters, verbose = _process_parameters(
        parameters, kw_parameters, resolved_udf.udf, type, _ignore_unknown_arguments
    )

    # Parsed cache TTL
    cache_max_age = _parse_ttl(cache_max_age, cache)

    dispatch_params = (
        type,
        resolved_udf.storage_type,
        engine,
        instance_type,
    )

    fn = _create_execution_function_async(
        dispatch_params,
        resolved_udf,
        x,
        y,
        z,
        cache_max_age,
        local_tile_bbox,
        _return_response,
        parameters,
        run_id,
    )

    if xyz_ignored and engine == "local":
        # This warning doesn't matter on realtime because we will just put the x/y/z into the URL
        warnings.warn(
            FusedIgnoredWarning(
                "x, y, z arguments will be ignored because the following packages were not all found: mercantile shapely geopandas"
            ),
        )

    return await _execute_with_retry_async(
        fn, max_retry, _cancel_callback, _return_response, verbose
    )



from typing import List

from fused._formatter.udf import fused_registry_repr


class Registry(dict):
    def _repr_html_(self):
        return fused_registry_repr(self)

    def __getattribute__(self, key):
        try:
            return super().__getattribute__(key)
        except AttributeError:
            try:
                return self[key]
            # Note that we need to raise an AttributeError, **not a KeyError** so that
            # IPython's _repr_html_ works here
            except KeyError:
                raise AttributeError(
                    f"object of type {type(self).__name__} has no attribute {key}"
                ) from None

    def __dir__(self) -> List[str]:
        """Provide method name lookup and completion. Only provide 'public'
        methods.
        """
        # This enables autocompletion
        extra_attrs = set(self.keys())
        normal_dir = {name for name in dir(type(self)) if not name.startswith("_")}

        return sorted(normal_dir | extra_attrs)



from __future__ import annotations

import contextlib
import sys
from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fused.models.udf import AnyBaseUdf


noop_decorators: ContextVar[bool] = ContextVar("noop_decorator", default=False)
decorator_src_override: ContextVar[str | None] = ContextVar(
    "decorator_src_override", default=None
)
decorator_udf_override: ContextVar[AnyBaseUdf | None] = ContextVar(
    "decorator_udf_override", default=None
)

decorator_import_globals_disabled: ContextVar[bool | None] = ContextVar(
    "decorator_import_globals_override", default=None
)


@contextmanager
def noop_decorators_context(val: bool):
    token = noop_decorators.set(val)
    try:
        yield token
    finally:
        noop_decorators.reset(token)


# Provides the UDF's code to the fused.udf decorator via the global context (above).
#
# Injecting the UDF code back inside the run function.
@contextmanager
def decorator_src_override_context(val: str):
    token = decorator_src_override.set(val)
    try:
        yield token
    finally:
        decorator_src_override.reset(token)


@contextmanager
def decorator_udf_override_context(udf: AnyBaseUdf):
    token = decorator_udf_override.set(udf)
    try:
        yield token
    finally:
        decorator_udf_override.reset(token)


@contextmanager
def decorator_import_globals_disable_context():
    token = decorator_import_globals_disabled.set(True)
    try:
        yield token
    finally:
        decorator_import_globals_disabled.reset(token)


class StdStreamProxy:
    """Act as a proxy for a target stream to prevent it from being closed unexpectedly"""

    def __init__(self, target: str):
        self._target = target
        self._stream = getattr(sys, target)

    # Forward all non-overridden methods
    def __getattr__(self, item):
        return getattr(self._stream, item)

    # To mitigate UDF crashes, we prevent close from operating on the stream. We ensure that anything left in the buffer
    # is flushed.
    def close(self):
        self.flush()

    def get_stream(self):
        return self._stream

    def __enter__(self):
        setattr(sys, self._target, self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        setattr(sys, self._target, self._stream)


@contextlib.contextmanager
def _isolate_streams():
    # We proxy all streams, including the dunder variants, to ensure none are closed within UDF code
    with (
        StdStreamProxy("stdin") as stdin_proxy,
        StdStreamProxy("__stdin__") as __stdin_proxy,
        StdStreamProxy("stdout") as stdout_proxy,
        StdStreamProxy("__stdout__") as __stdout_proxy,
        StdStreamProxy("stderr") as stderr_proxy,
        StdStreamProxy("__stderr__") as __stderr_proxy,
    ):
        yield (
            stdin_proxy,
            __stdin_proxy,
            stdout_proxy,
            __stdout_proxy,
            stderr_proxy,
            __stderr_proxy,
        )



import ast
import datetime
import time
from contextlib import ExitStack
from typing import Any

import fused
from fused._options import options as OPTIONS
from fused._udf.compile_v2 import compile_udf_and_run_with_kwargs
from fused._udf.state import (
    _isolate_streams,
    decorator_src_override_context,
)
from fused.models.udf import EMPTY_UDF, AnyBaseUdf
from fused.models.udf._eval_result import UdfEvaluationResult


def execute_against_sample(
    udf: AnyBaseUdf,
    input: list[Any],
    validate_imports: bool | None = None,
    _return_response: bool = False,
    cache_max_age: int | None = None,
    **kwargs,
) -> UdfEvaluationResult:
    if udf is EMPTY_UDF:
        raise ValueError("Empty UDF cannot be evaluated. Use `set_udf` to set a UDF.")

    # Validate import statements correspond to valid modules
    validate_imports_whitelist(udf, validate_imports=validate_imports)

    if not OPTIONS.local_engine_cache:
        cache_max_age = 0

    if cache_max_age is None:
        if udf.cache_max_age is not None:
            cache_max_age = udf.cache_max_age
        else:
            cache_max_age = int(datetime.timedelta(days=90).total_seconds())

    # Run UDF
    # TODO capture output
    exception_raised = None
    _output = None
    has_exception = False
    errormsg = None
    exception_class = None
    time_start = time.perf_counter()
    try:
        with _isolate_streams():
            if cache_max_age == 0:
                _output = compile_udf_and_run_with_kwargs(udf, *input, **kwargs)
            else:
                wrapped = fused.cache(
                    compile_udf_and_run_with_kwargs,
                    cache_max_age=cache_max_age,
                    cache_verbose=False,
                )
                _output = wrapped(udf, *input, **kwargs)
    except Exception as exc:
        exception_raised = exc
        has_exception = True
        exception_class = exc.__class__.__name__
        errormsg = f"{exception_class}: {str(exc)}"
        # TODO proper error traceback
        # traceback.print_tb(exc.__traceback__, file=err_buf)
    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    if _return_response:
        return UdfEvaluationResult(
            data=_output,
            udf=udf,
            time_taken_seconds=time_taken_seconds,
            stdout=None,
            stderr=None,
            error_message=errormsg,
            has_exception=has_exception,
            exception_class=exception_class,
        )
    if exception_raised:
        raise exception_raised
    return _output


def execute_for_decorator(udf: AnyBaseUdf) -> AnyBaseUdf:
    """Evaluate a UDF for the purpose of getting the UDF object out of it."""
    # Define custom function in environment

    # This is a stripped-down version of execute_against_sample, above.

    src = udf.code

    exec_globals_locals = {"fused": fused}

    with ExitStack() as stack:
        stack.enter_context(decorator_src_override_context(src))

        # Add headers to sys.meta_path
        if udf.headers is not None:
            for header in udf.headers:
                stack.enter_context(header._register_custom_finder())

        exec(src, exec_globals_locals, exec_globals_locals)

        if udf.entrypoint not in exec_globals_locals:
            raise NameError(
                f"Could not find {udf.entrypoint}. You need to define a UDF with `def {udf.entrypoint}()`."
            )

        _fn = exec_globals_locals[udf.entrypoint]

        return _fn


def validate_imports_whitelist(udf: AnyBaseUdf, validate_imports: bool | None = None):
    # Skip import validation if the option is set
    if not fused.options.default_validate_imports and validate_imports is not True:
        return

    # Skip import validation if not logged in
    if not fused.api.AUTHORIZATION.is_configured():
        return

    from fused._global_api import get_api

    # Get the dependency whitelist from the cached API endpoint
    api = get_api()
    package_dependencies = api.dependency_whitelist()

    # Initialize a list to store the import statements
    import_statements = []

    # Parse the source code into an AST
    tree = ast.parse(udf.code)

    # Traverse the AST to find import statements
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                import_statements.append(alias.name)
        elif isinstance(node, ast.ImportFrom):
            module_name = node.module
            import_statements.append(module_name)

    # Check for unavailable modules
    header_modules = [header.module_name for header in udf.headers]
    fused_modules = ["fused"]  # assume fused is always available
    available_modules = (
        list(package_dependencies["dependency_whitelist"].keys())
        + header_modules
        + fused_modules
    )
    unavailable_modules = []
    for import_statement in import_statements:
        if import_statement.split(".", 1)[0] not in available_modules:
            unavailable_modules.append(import_statement)

    if unavailable_modules:
        raise ValueError(
            f"The following imports in the UDF might not be available: {repr(unavailable_modules)}. Please check the UDF headers and imports and try again."
        )

    # TODO: check major versions for some packages



from __future__ import annotations

from contextlib import contextmanager
from http import HTTPStatus

from aiohttp import ClientResponse, ClientResponseError
from requests import HTTPError, Response, Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from fused._options import options as OPTIONS


def raise_for_status(r: Response):
    """
    A wrapper around Response.raise_for_status to give more information on 422.
    """
    # Look for an error message from model parsing
    if r.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        error_msg = r.json()
        raise HTTPError(error_msg, response=r)

    # Look for a Fused error message
    if 400 <= r.status_code < 600:
        to_raise: Exception | None = None
        try:
            if r.headers["Content-Type"] == "application/json":
                error_msg = r.json()
                if "message" in error_msg:
                    error_msg = error_msg["message"]
                to_raise = HTTPError(error_msg, response=r)
        except:  # noqa: E722
            # Fall through to raise_for_status, below
            pass
        if to_raise is not None:
            raise to_raise

    r.raise_for_status()


async def raise_for_status_async(r: ClientResponse):
    # Look for an error message from model parsing

    if r.status == HTTPStatus.UNPROCESSABLE_ENTITY:
        error_msg = await r.json()
        raise ClientResponseError(
            request_info=r.request_info,
            status=r.status,
            message=error_msg,
            headers=r.headers,
            history=r.history,
        )

    # Look for a Fused error message
    if 400 <= r.status < 600:
        to_raise: Exception | None = None
        try:
            if r.headers["Content-Type"] == "application/json":
                error_msg = await r.json()
                if "message" in error_msg:
                    error_msg = error_msg["message"]
                to_raise = ClientResponseError(
                    request_info=r.request_info,
                    status=r.status,
                    message=error_msg,
                    headers=r.headers,
                    history=r.history,
                )
        except:  # noqa: E722
            # Fall through to raise_for_status, below
            pass
        if to_raise is not None:
            raise to_raise

    r.raise_for_status()


default_retry_strategy = Retry(
    total=OPTIONS.request_max_retries,
    status_forcelist=[502, 503],
    backoff_factor=OPTIONS.request_retry_base_delay,
    allowed_methods=["HEAD", "GET", "OPTIONS", "PUT", "POST", "DELETE", "PATCH"],
)


@contextmanager
def session_with_retries(
    total_retries: int | None = None,
    backoff_factor: float | None = None,
):
    """
    Context manager that provides a Session with retry logic.

    :param total_retries: Total number of retries to allow.
    :param backoff_factor: A backoff factor (waiting time) to apply between attempts.
    """
    total_retries = total_retries or OPTIONS.request_max_retries
    backoff_factor = backoff_factor or OPTIONS.request_retry_base_delay

    session = Session()
    retry_strategy = Retry(
        total=total_retries,
        status_forcelist=[502, 503],
        backoff_factor=backoff_factor,
        allowed_methods=["HEAD", "GET", "OPTIONS", "PUT", "POST", "DELETE", "PATCH"],
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)
    session.mount("http://", adapter)
    session.mount("https://", adapter)

    try:
        yield session
    finally:
        session.close()



"""Vendored from xarray
https://github.com/pydata/xarray/blob/f4e0523be0ce8babaa8eff38365e5308b1fdb76b/xarray/core/formatting_html.py
"""

from __future__ import annotations

import json
import uuid
from functools import partial
from html import escape
from typing import TYPE_CHECKING, Any, Callable, Sequence

from pydantic import BaseModel

from fused._formatter.common import copyable_text, load_static_files
from fused._formatter.udf import udf_section

if TYPE_CHECKING:
    from fused.models.api import (
        JobConfig,
        JobStepConfig,
        PartitionJobStepConfig,
        UdfJobStepConfig,
    )
    from fused.models.udf import AnyBaseUdf


def format_dims(dims, dims_with_index):
    if not dims:
        return ""

    dim_css_map = {
        dim: " class='xr-has-index'" if dim in dims_with_index else "" for dim in dims
    }

    dims_li = "".join(
        f"<li><span{dim_css_map[dim]}>{escape(str(dim))}</span>: {size}</li>"
        for dim, size in dims.items()
    )

    return f"<ul class='xr-dim-list'>{dims_li}</ul>"


def collapsible_section(
    name,
    inline_details: str = "",
    details: str = "",
    n_items: int | None = None,
    enabled: bool = True,
    collapsed: bool = False,
):
    # "unique" id to expand/collapse the section
    data_id = "section-" + str(uuid.uuid4())

    has_items = n_items is not None and n_items
    n_items_span = "" if n_items is None else f" <span>({n_items})</span>"
    enabled_str = "" if enabled and has_items else "disabled"
    collapsed_str = "" if collapsed or not has_items else "checked"
    tip = " title='Expand/collapse section'" if enabled else ""

    return (
        f"<input id='{data_id}' class='xr-section-summary-in' "
        f"type='checkbox' {enabled_str} {collapsed_str}>"
        f"<label for='{data_id}' class='xr-section-summary' {tip}>"
        f"{name}:{n_items_span}</label>"
        f"<div class='xr-section-inline-details'>{inline_details}</div>"
        f"<div class='xr-section-details'>{details}</div>"
    )


def _mapping_section(
    mapping,
    name: str,
    details_func: Callable[[Any], str],
    max_items_collapse,
    enabled: bool = True,
    **kwargs,
):
    n_items = len(mapping)
    expanded = n_items < max_items_collapse
    collapsed = not expanded

    return collapsible_section(
        name,
        details=details_func(mapping, **kwargs),
        n_items=n_items,
        enabled=enabled,
        collapsed=collapsed,
    )


def _obj_repr(obj, header_components, sections, sections_class="xr-sections"):
    """Return HTML repr of an xarray object.

    If CSS is not injected (untrusted notebook), fallback to the plain text repr.

    """
    header = f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    sections = "".join(
        f"<li class='xr-section-item'>{s}</li>" for s in sections if s is not None
    )

    icons_svg, css_style = load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' style='display:none'>"
        f"{header}"
        f"<ul class='{sections_class}'>{sections}</ul>"
        "</div>"
        "</div>"
    )


def summarize_ingest_input(
    input: str,
    *,
    is_index: bool = False,
):
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    name = escape(str(input))

    return (
        f'<div class="xr-var-name" style="grid-column: 1 / -2;"><span{cssclass_idx}>{name}</span></div>'
        f"{copyable_text(name, show_text=False)}"
    )


def summarize_ingest_inputs(inputs: list[str]):
    li_items = []
    for input in inputs:
        li_content = summarize_ingest_input(input, is_index=False)
        li_items.append(f"<li class='xr-var-item'>{li_content}</li>")

    vars_li = "".join(li_items)

    return f'<ul class="xr-var-list">{vars_li}</ul>'


ingest_inputs_section = partial(
    _mapping_section,
    name="Inputs",
    details_func=summarize_ingest_inputs,
    max_items_collapse=25,
)


def get_object_settings(obj: BaseModel, exclude: Sequence[str] = ()) -> str:
    def _format_kv(key: str, val: Any) -> str:
        if key in exclude:
            return ""
        return f"<li>{key}: <code>{val}</code></li>"

    return f"<ul>{''.join([_format_kv(key, val) for key, val in obj.model_dump().items()])}</ul>"


def _collect_udfs(
    config: JobStepConfig, *, names: Sequence[str] = ("udf",)
) -> dict[str, AnyBaseUdf]:
    results = {}
    for udf_name in names:
        if hasattr(config, udf_name):
            udf_attr = getattr(config, udf_name)
            if udf_attr is not None:
                results[udf_attr.name] = udf_attr
    return results


def fused_ingestion_repr(ingest: PartitionJobStepConfig) -> str:
    obj_type = f"fused.{ingest.__class__.__name__}: {ingest.name or ingest.output or ingest.output_metadata}"
    header_components = [
        f"<div class='xr-obj-type'>{escape(obj_type)}</div>",
        get_object_settings(ingest, exclude=["input", "type"]),
    ]

    inputs = ingest.input if isinstance(ingest.input, list) else [ingest.input]

    sections: list[str] = [
        ingest_inputs_section(inputs),
    ]

    return _obj_repr(ingest, header_components, sections)


def fused_udf_step_repr(step: UdfJobStepConfig) -> str:
    name = step.name
    if not name and step.udf is not None:
        name = step.udf.name
    obj_type = f"fused.{step.__class__.__name__}: {name}"
    header_components = [
        f"<div class='xr-obj-type'>{escape(obj_type)}</div>",
        get_object_settings(step, exclude=["input", "udf", "type"]),
    ]

    inputs = [json.dumps(i) for i in step.input] if step.input is not None else []
    udfs = _collect_udfs(step)

    sections: list[str] = [
        ingest_inputs_section(inputs) if step.input is not None else None,
        udf_section(udfs),
    ]

    return _obj_repr(step, header_components, sections)


def _get_job_repr(i: int, step: JobStepConfig) -> str:
    if hasattr(step, "_repr_html_"):
        html = step._repr_html_()
    else:
        html = repr(step)
    return f'<li><h4>Step {i}</h4><div style="padding-left: 2em;">{html}</div></li>'


def fused_job_repr(job: JobConfig) -> str:
    return f"""
    <div class='xr-obj-type'>fused.JobConfig: {job.name}</div>
    <ul>
    {"".join([_get_job_repr(step_idx, step) for step_idx, step in enumerate(job.steps)])}
    </ul>
    """



from __future__ import annotations

import base64
import hashlib
import json
import os
import secrets
import socket
import webbrowser
from datetime import datetime, timedelta, timezone
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from textwrap import dedent
from threading import RLock
from urllib.parse import parse_qs, urlencode, urlparse

import requests
from loguru import logger
from pydantic import BaseModel, ValidationError
from typing_extensions import override

from fused._environment import is_pyodide
from fused._options import options as OPTIONS

_refresh_token_lock = RLock()


def _credentials_path():
    return Path(OPTIONS.auth.credentials_path).expanduser()


class BadCredentialsException(Exception):
    def __init__(self, message=""):
        self.http_status_code = 401
        self.detail = "Bad credentials"
        if message:
            self.detail = f"{self.detail}: {message}"


class Credentials(BaseModel):
    """A dataclass representation of OAuth2 credentials"""

    access_token: str
    auth_scheme: str
    refresh_token: str
    id_token: str
    scope: str
    expires_in: int
    expires_at: datetime

    @classmethod
    def authenticate(cls) -> Credentials:
        if OPTIONS.no_login:
            raise BadCredentialsException(
                "User has explicitly disabled automatic authentication using `fused.options.no_login`"
            )
        token_dict = authenticate()
        credentials = cls(**token_dict)
        credentials.save_to_disk()
        return credentials

    @classmethod
    def from_disk(cls) -> Credentials:
        path = _credentials_path()
        if path.exists():
            with open(path) as file:
                content = file.read()
            try:
                return cls.model_validate_json(content)
            except ValidationError as e:
                raise ValueError(f"Invalid credentials file: {e}")
        raise ValueError("Credentials file does not exist")

    @classmethod
    def from_disk_or_authenticate(cls) -> Credentials:
        try:
            return cls.from_disk()
        except ValueError:
            return cls.authenticate()

    def save_to_disk(self) -> None:
        if OPTIONS.save_user_settings:
            path = _credentials_path()
            path.parent.mkdir(parents=True, exist_ok=True)
            with open(path, "w") as f:
                f.write(self.model_dump_json())

    @classmethod
    def from_token_response(cls, token_dict: dict) -> Credentials:
        # Add an expires_at field in the dictionary for next time it's loaded
        token_dict["expires_at"] = (
            datetime.now(timezone.utc) + timedelta(seconds=token_dict["expires_in"] - 1)
        ).isoformat()
        token_dict["auth_scheme"] = "Bearer"

        credentials = cls(**token_dict)
        credentials.save_to_disk()

        return credentials

    def refresh_if_needed(self) -> Credentials:
        # Do not call the refresh token from multiple threads at once. If this happens,
        # the rate limit can accidentally be exceeded.
        # This could be optimized with double-checked locking, but the Wikipedia page
        # points out there could be an issue with the object being in a partially updated
        # state.
        # https://en.wikipedia.org/wiki/Double-checked_locking
        with _refresh_token_lock:
            if self.expires_at < datetime.now(timezone.utc):
                # Credentials are expired; we need to refresh
                token_dict = refresh_token(self.refresh_token)
                new_credentials = Credentials(**token_dict)
                self.__dict__.update(new_credentials.__dict__)

        return self


class BearerAccessToken:
    def __init__(self, access_token: str, auth_scheme: str = "Bearer"):
        self.access_token = access_token
        self.auth_scheme = auth_scheme


class MaybeInitializedAuthorization:
    """OAuth2 credentials or Bearer access token obtained from the global
    context that may or may not have been initialized.
    """

    _credentials: Credentials | None
    _bearer_access_token: BearerAccessToken | None

    def __init__(self) -> None:
        self._credentials = None
        if (token := self._maybe_retrieve_auth_token_from_global_context()) is not None:
            self._bearer_access_token = token
            return

        try:
            self._credentials = Credentials.from_disk()
        except ValueError:
            if OPTIONS.prompt_to_login:
                auth_msg = """\
                    Credentials not found on disk. Authenticate with:

                    from fused import NotebookCredentials
                    credentials = NotebookCredentials()
                """
                print(dedent(auth_msg))

    def reset(self) -> None:
        """Reset the loaded credentials"""
        self._credentials = None

    def initialize(self) -> None:
        """Force initialization of credentials."""
        _ = self.credentials

    def _maybe_retrieve_auth_token_from_global_context(
        self,
    ) -> BearerAccessToken | None:
        from fused.core._impl._context_impl import (
            context_get_auth_scheme_and_token,
            context_in_batch,
            context_in_realtime,
        )

        # SECURITY: We are currently inside RT2 and access to RT2 is already
        # authenticated via fused-server. But RT2 may send requests back to
        # fused-server for invocations of fused.run within a UDF. And for
        # fused-server to be able to authorize the request correctly, we need to
        # ensure that RT2 is including the access token back in the request.
        #
        # As a result, directly injecting the token back without additional
        # validation is acceptable here.
        maybe_auth_from_context = context_get_auth_scheme_and_token()
        if (
            maybe_auth_from_context
            and maybe_auth_from_context[0]
            and maybe_auth_from_context[1]
        ):
            return BearerAccessToken(
                maybe_auth_from_context[1], maybe_auth_from_context[0]
            )

        # Fallback to using token from env if the global context does not have
        # the token. But also ensure we only accept tokens of type Fused-Rtenv-Bearer.
        auth_scheme = os.getenv("FUSED_AUTH_TOKEN_TYPE", "")
        token = os.getenv("FUSED_AUTH_TOKEN", "")
        if (
            auth_scheme == "Fused-Rtenv-Bearer" or auth_scheme == "Fused-Env-Bearer"
        ) and token != "":
            return BearerAccessToken(token, auth_scheme)

        if context_in_realtime() or context_in_batch():
            raise BadCredentialsException("AuthN configuration error.")

    @property
    def credentials(self) -> Credentials | BearerAccessToken:
        """
        Retrieve valid credentials, initializing them from global context or
        authenticating from scratch if needed.
        """
        if (token := self._maybe_retrieve_auth_token_from_global_context()) is not None:
            self._bearer_access_token = token
            return self._bearer_access_token

        existing_credentials = (
            self._credentials or Credentials.from_disk_or_authenticate()
        )

        existing_credentials = existing_credentials.refresh_if_needed()
        self._credentials = existing_credentials
        return existing_credentials

    def set_credentials(self, credentials: Credentials) -> None:
        self._credentials = credentials

    def is_configured(self) -> bool:
        try:
            if self._maybe_retrieve_auth_token_from_global_context() is not None:
                return True

            if OPTIONS.no_login:
                return False

            if self._credentials is not None:
                return True

            return Credentials.from_disk() is not None
        except Exception:
            return False


AUTHORIZATION: MaybeInitializedAuthorization = MaybeInitializedAuthorization()
"""Global credentials."""


def _find_available_port(start_port: int) -> int:
    """
    Find first available port to be used for the temporary authorization server.
    """
    server_socket = None
    end_port = start_port + 3
    # the ports must be whitelisted in auth0 before being used here
    while start_port <= end_port:
        try:
            # Try to bind a temporary socket to check availability
            server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # SO_REUSEADDR allows quickly rebinding to the same port after closing
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind(("localhost", start_port))
            return start_port
        except OSError:
            logger.warning(f"Port {start_port} is not available, trying the next port.")
            start_port += 1
        finally:
            if server_socket:
                server_socket.close()

    raise OSError(
        f"Could not find an available port between {start_port} and {end_port}."
    )


# TODO: remove some of these functions below
def authenticate():
    if is_pyodide():
        # We cannot use this method to log in in Pyodide, so show an error message instead.
        raise PermissionError(
            "Fused credentials not found -- please log in using the Workbench to continue."
        )

    print(
        'Existing credentials not found on disk: Please check your browser to log in. If using a notebook, use "credentials = fused.api.NotebookCredentials()" to log in.'
    )
    code_verifier = secrets.token_urlsafe(48)
    code_challenge = get_code_challenge(code_verifier)

    callback_server_port = _find_available_port(3000)
    redirect_uri = f"http://localhost:{callback_server_port}"

    params = {
        "audience": OPTIONS.auth.audience,
        "scope": " ".join(OPTIONS.auth.scopes),
        "response_type": "code",
        "client_id": OPTIONS.auth.client_id,
        "redirect_uri": redirect_uri,
        "code_challenge_method": "S256",
        "code_challenge": code_challenge,
    }
    authorize_url = f"{OPTIONS.auth.authorize_url}?{urlencode(params)}"
    code = handle_redirect(authorize_url, callback_server_port)

    token_data = {
        "client_id": OPTIONS.auth.client_id,
        "grant_type": "authorization_code",
        "audience": OPTIONS.auth.audience,
        "client_secret": OPTIONS.auth.client_secret,
        "code": code,
        "redirect_uri": redirect_uri,
        "code_verifier": code_verifier,
    }
    token_response = requests.post(
        OPTIONS.auth.oauth_token_url, json=token_data, timeout=OPTIONS.request_timeout
    )
    token_response.raise_for_status()
    token_dict = token_response.json()

    # Add an expires_at field in the dictionary for next time it's loaded
    token_dict["expires_at"] = (
        datetime.now(timezone.utc) + timedelta(seconds=token_dict["expires_in"] - 1)
    ).isoformat()
    token_dict["auth_scheme"] = "Bearer"

    save_token_to_disk(token_dict)
    return token_dict


def logout():
    """Open the user's browser to the Auth0 logout page."""
    webbrowser.open(OPTIONS.auth.logout_url)


def save_token_to_disk(token: dict):
    if OPTIONS.save_user_settings:
        path = _credentials_path()
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w") as f:
            json.dump(token, f)


def delete_token_from_disk():
    _credentials_path().unlink(missing_ok=True)


def get_code_challenge(code_verifier: str) -> str:
    """Take an input string and hash it to generate a challenge string

    Refer to https://auth0.com/docs/get-started/authentication-and-authorization-flow/call-your-api-using-the-authorization-code-flow-with-pkce
    """
    code_challenge_digest = hashlib.sha256(code_verifier.encode("utf-8")).digest()
    code_challenge = base64.urlsafe_b64encode(code_challenge_digest).decode("utf-8")
    return code_challenge.replace("=", "")


def handle_redirect(authorize_url: str, port: int) -> str:
    """Open the authorization url and intercept its redirect

    The redirection from the `/authorize` endpoint includes a code that can be used
    against the `/oauth/token` endpoint to fetch a refresh and access token.
    """
    code: str | None = None

    class RequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            nonlocal code

            query_string = urlparse(self.path).query
            parsed_qs = parse_qs(query_string)

            assert len(parsed_qs["code"]) == 1
            code = parsed_qs["code"][0]

            self.send_response(200)
            self.send_header("Content-Type", "text/html")
            self.end_headers()

            text = "Success! You can now close this tab and return to Python."
            self.wfile.write(text.encode())

        def log_message(self, format, *args):
            # This is overwritten to prevent the server from logging requests to the
            # console
            return

    class HTTPServerWithTimeout(HTTPServer):
        timeout: float | None = 30

        @override
        def handle_timeout(self) -> None:
            # Shutdown the server before raising the exception
            self.server_close()
            raise TimeoutError(
                f"Authentication timed out after {self.timeout} seconds. Please "
                "try again and complete the browser authentication flow."
            )

    # Use the provided dynamic port
    server = HTTPServerWithTimeout(("localhost", port), RequestHandler)

    webbrowser.open(authorize_url)
    # Note that this only handles _one_ request, but that's all it should need
    server.handle_request()
    # Shutdown the server after done
    server.server_close()

    assert code is not None
    return code


def refresh_token(refresh_token: str):
    """Generate a new access_token using a refresh token"""
    token_data = {
        "client_id": OPTIONS.auth.client_id,
        "grant_type": "refresh_token",
        "client_secret": OPTIONS.auth.client_secret,
        # TODO: Is this needed? This won't be right for notebook
        "redirect_uri": OPTIONS.auth.local_redirect_url,
        "refresh_token": refresh_token,
    }
    token_response = requests.post(
        OPTIONS.auth.oauth_token_url,
        json=token_data,
        timeout=OPTIONS.request_timeout,
    )
    token_response.raise_for_status()
    token_dict = token_response.json()

    # Add an expires_at field in the dictionary for next time it's loaded
    token_dict["expires_at"] = (
        datetime.now(timezone.utc) + timedelta(seconds=token_dict["expires_in"] - 1)
    ).isoformat()
    token_dict["refresh_token"] = refresh_token
    token_dict["auth_scheme"] = "Bearer"

    save_token_to_disk(token_dict)
    return token_dict



import os
from enum import Enum
from typing import Optional

from fused._options import ShowOptions
from fused._options import options as OPTIONS


class Environment(str, Enum):
    AZURE = "azure"
    COCALC = "cocalc"
    COLAB = "colab"
    DATABRICKS = "databricks"
    KAGGLE = "kaggle"
    NTERACT = "nteract"
    OTHER_IPYTHON = "other_ipython"
    TERMINAL = "terminal"
    UNKNOWN = "unknown"
    VSCODE = "vscode"


def deduce_environment() -> Environment:
    # https://github.com/plotly/plotly.py/blob/fc3ef002c7a7898b9244b1549757a64e8df266dd/packages/python/plotly/plotly/io/_renderers.py#L469

    try:
        import IPython

        ipython = IPython.get_ipython()
        try:
            import google.colab  # noqa: F401

            return Environment.COLAB
        except ImportError:
            pass

        # Check if we're not in a notebook
        if ipython is None:
            return Environment.UNKNOWN

        # Check if we're running in a Kaggle notebook
        if os.path.exists("/kaggle/input"):
            return Environment.KAGGLE

        # Check if we're running in an Azure Notebook
        if "AZURE_NOTEBOOKS_HOST" in os.environ:
            return Environment.AZURE

        # Check if we're running in VSCode
        if "VSCODE_PID" in os.environ:
            return Environment.VSCODE

        # Check if we're running in nteract
        if "NTERACT_EXE" in os.environ:
            return Environment.NTERACT

        # Check if we're running in CoCalc
        if "COCALC_PROJECT_ID" in os.environ:
            return Environment.COCALC

        if "DATABRICKS_RUNTIME_VERSION" in os.environ:
            return Environment.DATABRICKS

        if ipython.__class__.__name__ == "TerminalInteractiveShell":
            return Environment.TERMINAL

        return Environment.OTHER_IPYTHON
    except ImportError:
        pass

    return Environment.UNKNOWN


def infer_display_method(open_browser: Optional[bool], show_widget: Optional[bool]):
    # Passed-in options have highest precedence
    if open_browser is not None or show_widget is not None:
        return ShowOptions(open_browser=open_browser, show_widget=show_widget)

    if OPTIONS.show.open_browser is not None or OPTIONS.show.show_widget is not None:
        return OPTIONS.show

    if ENVIRONMENT in [
        Environment.TERMINAL,
        Environment.UNKNOWN,
    ]:
        return ShowOptions(open_browser=True, show_widget=False)
    else:
        return ShowOptions(open_browser=False, show_widget=True)


def is_pyodide():
    # https://pyodide.org/en/stable/usage/faq.html#how-to-detect-that-code-is-run-with-pyodide
    import sys

    return "pyodide" in sys.modules


ENVIRONMENT = deduce_environment()



from typing import TYPE_CHECKING, Optional, TypeVar

if TYPE_CHECKING:
    import geopandas as gpd
    import mercantile
    import shapely

Tile = TypeVar("Tile", bound="gpd.GeoDataFrame")
Bounds = TypeVar("Bounds", bound=list)
TileXYZ = TypeVar("TileXYZ", bound="mercantile.Tile")
TileGDF = TypeVar("TileGDF", bound="gpd.GeoDataFrame")
ViewportGDF = TypeVar("ViewportGDF", bound="gpd.GeoDataFrame")
Bbox = TypeVar("Bbox", bound="shapely.geometry.polygon.Polygon")


class UdfRuntimeError(RuntimeError):
    def __init__(self, *args, child_exception_class: Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.child_exception_class = child_exception_class


class UdfSerializationError(Exception):
    pass


class UdfTimeoutError(RuntimeError):
    pass



from typing import Literal

from fused._options import options as OPTIONS
from fused._udf.context import get_global_context


def make_realtime_url(client_id: str | None) -> str:
    context = get_global_context()

    if client_id is None and context is not None:
        client_id = context.realtime_client_id

    if client_id is None:
        client_id = OPTIONS.realtime_client_id

    if client_id is None:
        raise ValueError(
            "Failed to detect realtime client ID (context is "
            "not configured with client ID)"
        )

    return f"{OPTIONS.base_url}/realtime/{client_id}"


def make_shared_realtime_url(id: str) -> str:
    return f"{OPTIONS.base_url}/realtime-shared/{id}"


def get_recursion_factor() -> int:
    context = get_global_context()

    recursion_factor = context.recursion_factor
    if recursion_factor is None:
        recursion_factor = 1
    if recursion_factor > OPTIONS.max_recursion_factor:
        raise ValueError(
            f"Recursion factor {recursion_factor} exceeds maximum {OPTIONS.max_recursion_factor}"
        )

    return recursion_factor + 1


def default_run_engine() -> Literal["remote", "local"]:
    if OPTIONS.default_udf_run_engine is not None:
        return OPTIONS.default_udf_run_engine
    return "remote"



import base64
from io import BytesIO
from typing import TYPE_CHECKING, Iterable, Optional, Union

if TYPE_CHECKING:
    import geopandas as gpd


def _normalize_url(url: str) -> str:
    if url.endswith("/"):
        return url[:-1]
    return url


def get_chunks_metadata(url: str) -> "gpd.GeoDataFrame":
    """Returns a GeoDataFrame with each chunk in the table as a row.

    Args:
        url: URL of the table.
    """
    import geopandas as gpd
    import pyarrow.parquet as pq
    import shapely

    url = _normalize_url(url) + "/_sample"

    # do not use pq.read_metadata as it may segfault in versions >= 12 (tested on 15.0.1)
    table = pq.read_table(url)
    metadata = _rewrite_metadata_to_target_version(table.schema.metadata)

    metadata_bytes = metadata[b"fused:_metadata"]
    fused_meta_buffer = base64.b64decode(metadata_bytes)
    with BytesIO(fused_meta_buffer) as bio:
        df = pq.read_table(bio).to_pandas()
    geoms = shapely.box(
        df["bbox_minx"], df["bbox_miny"], df["bbox_maxx"], df["bbox_maxy"]
    )
    return gpd.GeoDataFrame(df, geometry=geoms, crs="EPSG:4326")


def _rewrite_metadata_to_target_version(
    kvmeta: dict[bytes, bytes],
) -> dict[bytes, bytes]:
    """
    Rewrites the metadata to the target internal fused format version (v5).
    """
    format_version = kvmeta.get(b"fused:format_version", None)
    if format_version is None:
        raise ValueError("Dataset does not have Fused metadata.")

    if format_version == b"4":
        newmeta = {
            **kvmeta,
            b"fused:format_version": b"5",
        }

        metabytes = base64.decodebytes(kvmeta[b"fused:_metadata"])
        metabytes2 = metabytes[:20] + metabytes[20 + 20 :]
        metabytes3 = base64.encodebytes(metabytes2)

        newmeta[b"fused:_metadata"] = metabytes3
        return newmeta

    elif format_version == b"5":
        return kvmeta

    else:
        raise ValueError("Dataset has an incompatible metadata version.")


def get_chunk_from_table(
    url: str,
    file_id: Union[str, int, None],
    chunk_id: Optional[int],
    *,
    columns: Optional[Iterable[str]] = None,
) -> "gpd.GeoDataFrame":
    """Returns a chunk from a table and chunk coordinates.

    This can be called with file_id and chunk_id from `get_chunks_metadata`.

    Args:
        url: URL of the table.
        file_id: File ID to read.
        chunk_id: Chunk ID to read.
        columns: Read only the specified columns.
    """
    import geopandas as gpd
    import geopandas.io.arrow
    import pyarrow.parquet as pq

    if file_id is None:
        data = gpd.read_parquet(url)
    else:
        url = _normalize_url(url) + f"/{file_id}.parquet"

        with pq.ParquetFile(url) as file:
            if chunk_id is not None:
                table = file.read_row_group(chunk_id, columns=columns)
            else:
                table = file.read(columns=columns)

            if b"geo" in table.schema.metadata:
                data = geopandas.io.arrow._arrow_to_geopandas(table)
            else:
                data = table.to_pandas()

    if isinstance(data, (gpd.GeoDataFrame, gpd.GeoSeries)):
        if data.crs is None:
            data = data.set_crs(4326)
    return data



# ruff: noqa: F401, F403

from ._cache import cache, cache_call, cache_call_async

# create_path is deprecated
from ._download import create_path, download, file_path
from ._realtime_ops import (
    run_file,
    run_file_async,
    run_shared_file,
    run_shared_file_async,
    run_shared_tile,
    run_shared_tile_async,
    run_tile,
    run_tile_async,
)
from ._table_ops import get_chunk_from_table, get_chunks_metadata
from ._udf_ops import (
    load_udf_from_code,
    load_udf_from_directory,
    load_udf_from_file,
    load_udf_from_fused,
    load_udf_from_fused_async,
    load_udf_from_github,
    load_udf_from_shared_token,
    load_udf_from_zip,
)



"""
Models intended for use in the public API. These models hold a reference to FusedAPI for
easy chained operations.
"""

# ruff: noqa: F401

from ._cronjob import CronJob
from ._list import ListDetails
from .job import (
    AnyJobStepConfig,
    GeospatialPartitionJobStepConfig,
    JobConfig,
    JobStepConfig,
    NonGeospatialPartitionJobStepConfig,
    PartitionJobStepConfig,
    RootAnyJobStepConfig,
    UdfJobStepConfig,
)
from .udf_access_token import UdfAccessToken, UdfAccessTokenList



"""Vendored from xarray
https://github.com/pydata/xarray/blob/f4e0523be0ce8babaa8eff38365e5308b1fdb76b/xarray/core/formatting_html.py
"""

from __future__ import annotations

import uuid
from functools import partial
from html import escape
from typing import TYPE_CHECKING, Any, Callable

from fused._formatter.common import copyable_text, icon, load_static_files
from fused._formatter.noraise import noraise
from fused._formatter.udf import udf_section

if TYPE_CHECKING:
    from fused.models.udf import AnyBaseUdf
    from fused.models.udf._eval_result import UdfEvaluationResult


def _obj_repr(obj, header_components, sections):
    """Return HTML repr of an xarray object.

    If CSS is not injected (untrusted notebook), fallback to the plain text repr.

    """
    header = f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    sections = "".join(f"<li class='xr-section-item'>{s}</li>" for s in sections)

    icons_svg, css_style = load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' style='display:none'>"
        f"{header}"
        f"<ul class='xr-sections'>{sections}</ul>"
        "</div>"
        "</div>"
    )


def collapsible_section(
    name,
    inline_details: str = "",
    details: str = "",
    n_items: int | None = None,
    enabled: bool = True,
    collapsed: bool = False,
):
    # "unique" id to expand/collapse the section
    data_id = "section-" + str(uuid.uuid4())

    has_items = n_items is not None and n_items
    n_items_span = "" if n_items is None else f" <span>({n_items})</span>"
    enabled_str = "" if enabled and has_items else "disabled"
    collapsed_str = "" if collapsed or not has_items else "checked"
    tip = " title='Expand/collapse section'" if enabled else ""

    return (
        f"<input id='{data_id}' class='xr-section-summary-in' "
        f"type='checkbox' {enabled_str} {collapsed_str}>"
        f"<label for='{data_id}' class='xr-section-summary' {tip}>"
        f"{name}:{n_items_span}</label>"
        f"<div class='xr-section-inline-details'>{inline_details}</div>"
        f"<div class='xr-section-details'>{details}</div>"
    )


def _mapping_section(
    mapping,
    name: str,
    details_func: Callable[[Any], str],
    max_items_collapse,
    enabled: bool = True,
):
    n_items = len(mapping)
    expanded = n_items < max_items_collapse
    collapsed = not expanded

    return collapsible_section(
        name,
        details=details_func(mapping),
        n_items=n_items,
        enabled=enabled,
        collapsed=collapsed,
    )


def _render_decorator(udf: AnyBaseUdf) -> str | None:
    args: list[str] = []
    # Entrypoint isn't set here, because the decorator should be applied to a
    # specific function, and that function will be the entrypoint.
    # Name isn't set here because we are not indexing a registry by the UDF
    # name anymore.

    if udf.original_headers:
        args.append(f"headers={udf.original_headers}")

    return f"@fused.udf({', '.join(args)})"


@noraise(incompat_version_message="Log formatting failed", default="")
def summarize_log(
    log_name: str,
    log: str | None,
    *,
    is_index: bool = False,
    dtype=None,
):
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    log_lines = log.splitlines() if log is not None else []
    dims_str = f"lines:{len(log_lines)}"
    name = escape(log_name)
    preview = ""
    dtype = ""

    # "unique" ids required to expand/collapse subsections
    attrs_id = "attrs-" + str(uuid.uuid4())
    data_id = "data-" + str(uuid.uuid4())
    chunk_repr_disabled = "disabled" if not log else ""
    chunk_repr_checked = (
        "checked" if log_name == "stderr" and not chunk_repr_disabled else ""
    )

    # TODO: Fix for no schema, fix for parameters expected by the UDF, fix for copy decorator, fix for entrypoint
    attrs_ul = ""
    rendered_code = "\n".join(
        [f'<span class="fused-udf-body">{escape(line)}</span>' for line in log_lines]
    )
    udf_counter_style_name = f"fused-udf-code-number-style-{len(str(len(log_lines)))}"
    udf_counter_style = f"""
    <style>
    @counter-style {udf_counter_style_name} {{
        system: numeric;
        symbols: "0" "1" "2" "3" "4" "5" "6" "7" "8" "9";
        pad: {len(str(len(log_lines)))} " ";
    }}
    </style>
    """
    data_repr = f"""<div class="fused-udf-body-wrapper" style="--fused-udf-gutter-counter-style: {udf_counter_style_name};"><pre class="fused-udf-body"><code>{rendered_code}</code></pre></div>"""
    attrs_icon = icon("icon-file-text2")
    code_icon = icon("icon-code")

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dims_str}</div>"
        f"<div class='xr-var-dtype'>{dtype}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox' disabled>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<input id='{data_id}' class='xr-var-data-in' type='checkbox' {chunk_repr_disabled} {chunk_repr_checked}>"
        f"<label for='{data_id}' title='Show/Hide data repr'>"
        f"{code_icon}</label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
        f"<div class='xr-var-data'>{udf_counter_style}{data_repr}</div>"
    )


def summarize_logs(logs: dict[str, str | None]):
    li_items = []
    for log_name, log in logs.items():
        li_content = summarize_log(log_name, log, is_index=False)
        li_items.append(f"<li class='xr-var-item'>{li_content}</li>")

    vars_li = "".join(li_items)

    return f"<ul class='xr-var-list'>{vars_li}</ul>"


log_section = partial(
    _mapping_section,
    name="Logs",
    details_func=summarize_logs,
    max_items_collapse=25,
)


@noraise(incompat_version_message="Error formatting failed", default="")
def _summarize_error(result: UdfEvaluationResult):
    cssclass_idx = ""
    dims_str = ""
    name = "Error"
    preview = ""
    dtype = ""

    # "unique" ids required to expand/collapse subsections
    attrs_id = "attrs-" + str(uuid.uuid4())
    data_id = "data-" + str(uuid.uuid4())

    # TODO: Fix for no schema, fix for parameters expected by the UDF, fix for copy decorator, fix for entrypoint
    attrs_ul = f"""
    <pre>{escape(result.error_message)}</pre>
    """
    has_error = result.error_lineno is not None
    if has_error:
        chunk_repr_disabled = ""
        code_lines = result.udf.code.splitlines()
        ERROR_CONTEXT_LINE_COUNT = 10
        if ERROR_CONTEXT_LINE_COUNT is not None:
            min_lineno = result.error_lineno - ERROR_CONTEXT_LINE_COUNT
            max_lineno = result.error_lineno + ERROR_CONTEXT_LINE_COUNT
            start_lineno = max(0, min_lineno)
        else:
            min_lineno = 0
            max_lineno = len(code_lines)
            start_lineno = 0
        rendered_code = "\n".join(
            [
                f'<span class="fused-udf-body {"fused-udf-error-line" if lineno == result.error_lineno else ""}">{escape(line)}</span>'
                for lineno, line in enumerate(code_lines)
                if min_lineno < lineno < max_lineno
            ]
        )

        udf_counter_style_name = (
            f"fused-udf-code-number-style-{len(str(len(code_lines)))}"
        )
        udf_counter_style = f"""
        <style>
        @counter-style {udf_counter_style_name} {{
            system: numeric;
            symbols: "0" "1" "2" "3" "4" "5" "6" "7" "8" "9";
            pad: {len(str(len(code_lines)))} " ";
        }}
        </style>
        """

        data_repr = f"""<div class="fused-udf-body-wrapper" style="--fused-udf-gutter-counter-style: {udf_counter_style_name};"><pre class="fused-udf-body" style="counter-reset: udf-code-number {start_lineno};"><code>{rendered_code}</code></pre></div>"""
    else:
        chunk_repr_disabled = "disabled"
        data_repr = ""
        udf_counter_style = ""

    attrs_icon = icon("icon-file-text2")
    code_icon = icon("icon-code")

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dims_str}</div>"
        f"<div class='xr-var-dtype'>{dtype}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox' checked>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<input id='{data_id}' class='xr-var-data-in' type='checkbox' {chunk_repr_disabled} checked>"
        f"<label for='{data_id}' title='Show/Hide data repr'>"
        f"{code_icon}</label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
        f"<div class='xr-var-data'>{udf_counter_style}{data_repr}</div>"
    )


def summarize_error(results: list[UdfEvaluationResult]):
    assert len(results) == 1
    return f"<ul class='xr-var-list'><li class='xr-var-item'>{_summarize_error(results[0])}</li></ul>"


error_section = partial(
    _mapping_section,
    name="Error",
    details_func=summarize_error,
    max_items_collapse=25,
)


def fused_eval_result_repr(result: UdfEvaluationResult) -> str:
    data_repr = ""
    # TODO: Can't use dict for the schema because the types don't render in that case
    if not result.error_message:
        if result.data is not None and hasattr(result.data, "_repr_html_"):
            data_repr = result.data._repr_html_()
        else:
            data_repr = str(result.data)

    obj_type = f"Local result: {result.udf.name}" if result.udf else "Realtime result"
    header_components = [f"<div class='xr-obj-type'>{escape(obj_type)}</div>"]

    udfs = {result.udf.name: result.udf} if result.udf else {}

    sections: list[str] = [
        data_repr,
        udf_section(udfs),
    ]
    if result.stdout or result.stderr:
        sections.append(log_section({"stdout": result.stdout, "stderr": result.stderr}))
    if result.error_message or result.error_lineno:
        sections.append(error_section([result]))
    decorator_str = _render_decorator(result.udf) if result.udf else None
    schema_section = (
        f"<div>Decorator: <code>{copyable_text(decorator_str)}</code></div>"
        if not result.error_message and decorator_str
        else ""
    )
    time_taken_section = (
        f"<div>Time taken: <b>{result.time_taken_seconds:.4f}</b> seconds</div>"
    )
    return "".join(
        [
            _obj_repr(udfs, header_components, sections),
            schema_section,
            time_taken_section,
        ]
    )



from __future__ import annotations

import re
import warnings
from typing import Any, Literal, Sequence, Union
from urllib.parse import urlparse

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
    field_validator,
)
from typing_extensions import Annotated

from fused._formatter.formatter_job_config import (
    fused_ingestion_repr,
    fused_job_repr,
    fused_udf_step_repr,
)
from fused._str_utils import is_url
from fused.models.udf._udf_registry import UdfRegistry
from fused.warnings import (
    FusedDefaultWarning,
    FusedIgnoredWarning,
    FusedPathWarning,
    FusedTypeWarning,
    FusedWarning,
)

from ...models.schema import Schema
from .._codegen import (
    create_directory_and_zip,
    extract_parameters,
    generate_meta_json,
    generate_readme,
    stringify_named_params,
    stringify_output,
    structure_params,
)
from .._inplace import _maybe_inplace
from ..base import FusedBaseModel, UserMetadataType
from ..internal import AnyDatasetOutput, DatasetOutputV2, RunResponse
from ..request import WHITELISTED_INSTANCE_TYPES
from ..udf import AnyBaseUdf, RootAnyBaseUdf
from ..udf._eval_result import MultiUdfEvaluationResult, UdfEvaluationResult

STR_IMPORTS = (
    "\n".join(
        [
            "import fused",
        ]
    )
    + "\n\n"
)

SYSTEM_PARAMETER_NAMES = {"output", "context"}
"""Parameter names that will be provided by the Fused system"""


def _common_validate_for_overwrite(
    table_name: str,
    table_names: Sequence[str],
    overwrite: bool = False,
):
    if table_name in table_names:
        if overwrite:
            warnings.warn(
                FusedDefaultWarning(
                    f"Table `{table_name}` already exists and will be overwritten because `overwrite=True`."
                )
            )
        else:
            raise ValueError(
                f"Table `{table_name}` already exists and cannot be written to."
            )


def _common_validate_for_run(
    *,
    output: AnyDatasetOutput | None,
    udf: AnyBaseUdf | None,
    ignore_no_udf: bool = False,
    ignore_no_output: bool = False,
    validate_imports: bool | None = None,
):
    """
    Validate that the job config is ready to run
    """
    if not ignore_no_udf and not udf.code:
        raise ValueError(
            f"No UDF code is set for `{udf.name}` UDF. Set the `udf` attribute or pass `ignore_no_udf=True`."
        )

    if not ignore_no_output and hasattr(output, "table") and not output.table:
        raise ValueError(
            f"No `table` attribute is set on the job output configuration for `{udf.name}` UDF. Set it to write output, or pass `ignore_no_output=True`"
        )

    if not ignore_no_output and hasattr(output, "url") and not output.url:
        raise ValueError(
            f"No `url` attribute is set on the job output configuration for `{udf.name}` UDF. Set it to write output, or pass `ignore_no_output=True`"
        )

    # TODO: Check if the output table name already exists

    # Check if headers are valid
    if not ignore_no_udf and not _validate_headers_for_remote_exec(udf):
        raise ValueError(f"Headers for UDF {udf.name} cannot resolve remotely.")

    # Validate import stamements correspond to valid modules
    if udf is not None:
        from fused._udf.execute_v2 import validate_imports_whitelist

        validate_imports_whitelist(udf, validate_imports=validate_imports)


def _assert_udf_has_parameters(udf: AnyBaseUdf):
    assert hasattr(udf, "set_parameters"), (
        f"Cannot set parameters on a UDF without parameters: {type(udf)}"
    )


class JobStepConfig(FusedBaseModel):
    type: Literal[
        "partition_geospatial",
        "partition_nongeospatial",
        "udf",
    ]
    name: StrictStr | None = None
    metadata: UserMetadataType = None
    """User defined metadata. Any metadata values must be JSON serializable."""
    _validate_version: bool = True
    ignore_chunk_error: bool = False
    """If `True`, continue processing even if some computations throw errors."""

    @property
    def input_metadata(self) -> DatasetInputBase | None:
        return None

    def _validate_for_run(
        self,
        *,
        ignore_no_udf: bool = False,
        ignore_no_output: bool = False,
        validate_imports: bool | None = None,
        validate_inputs: bool = True,
    ):
        has_udf = hasattr(self, "udf")
        has_output = hasattr(self, "output")
        _common_validate_for_run(
            output=self.output if has_output else None,
            udf=self.udf if has_udf else None,
            ignore_no_udf=ignore_no_udf if has_udf else True,
            ignore_no_output=ignore_no_output if has_output else True,
            validate_imports=validate_imports,
        )

    def run_batch(
        self,
        output_table: str | None = ...,
        instance_type: Literal["small", "medium", "large", WHITELISTED_INSTANCE_TYPES]
        | None = None,
        *,
        region: str | None = None,
        disk_size_gb: int | None = None,
        additional_env: list[str] | None = None,
        image_name: str | None = None,
        ignore_no_udf: bool = False,
        ignore_no_output: bool = False,
        validate_imports: bool | None = None,
        validate_inputs: bool = True,
        overwrite: bool | None = None,
        name: StrictStr | None = None,
        send_status_email: bool | None = None,
    ) -> RunResponse:
        """Execute this operation

        Args:
            output_table: The name of the table to write to. Defaults to None.
            instance_type: The AWS EC2 instance type to use for the job. Acceptable strings are "m5.large", "m5.xlarge", "m5.2xlarge", "m5.4xlarge", "m5.8xlarge", "m5.12xlarge", "m5.16xlarge", "r5.large", "r5.xlarge", "r5.2xlarge", "r5.4xlarge", "r5.8xlarge", "r5.12xlarge", or "r5.16xlarge". Defaults to None.
            region: The AWS region in which to run. Defaults to None.
            disk_size_gb: The disk size to specify for the job. Defaults to None.
            additional_env: Any additional environment variables to be passed into the job. Defaults to None.
            image_name: Custom image name to run. Defaults to None for default image.

            ignore_no_udf: Ignore validation errors about not specifying a UDF. Defaults to False.
            ignore_no_output: Ignore validation errors about not specifying output location. Defaults to False.
            name: Job name (optional). If not provided, the name will be generated automatically.
            send_status_email: Whether to send a status email to the user when the job is complete. Defaults to True.
        """
        to_run = self.model_copy(deep=True)

        if instance_type in ("small", "medium", "large"):
            instance_type = {
                "small": "t3.small",
                "medium": "m5.4xlarge",
                "large": "r5.16xlarge",
            }[instance_type]

        if output_table is Ellipsis:
            if hasattr(self, "output"):
                if not self.output.table:
                    raise ValueError("The Job requires `output_table` to be specified.")
                elif overwrite is not None:
                    to_run.output.overwrite = overwrite
        elif output_table is not None:
            to_run = to_run.set_output(
                table_or_url=output_table, overwrite=overwrite, inplace=False
            )
        else:
            ignore_no_output = True
        return to_run._run_remote(
            instance_type=instance_type,
            region=region,
            disk_size_gb=disk_size_gb,
            additional_env=additional_env,
            image_name=image_name,
            ignore_no_udf=ignore_no_udf,
            ignore_no_output=ignore_no_output,
            validate_imports=validate_imports,
            validate_inputs=validate_inputs,
            name=name,
            send_status_email=send_status_email,
        )

    run_remote = run_batch

    def _run_remote(
        self,
        instance_type: WHITELISTED_INSTANCE_TYPES | None = None,
        *,
        region: str | None = None,
        disk_size_gb: int | None = None,
        additional_env: list[str] | None = None,
        image_name: str | None = None,
        ignore_no_udf: bool = False,
        ignore_no_output: bool = False,
        validate_imports: bool | None = None,
        validate_inputs: bool = True,
        name: str | None = None,
        send_status_email: bool | None = None,
    ) -> RunResponse:
        # TODO: if the user uses start_job, this validation will never happen
        self._validate_for_run(
            ignore_no_udf=ignore_no_udf,
            ignore_no_output=ignore_no_output,
            validate_imports=validate_imports,
            validate_inputs=validate_inputs,
        )
        config = JobConfig(steps=[self])
        if name is not None:
            config.name = name

        return self._api.start_job(
            config,
            instance_type=instance_type,
            region=region,
            disk_size_gb=disk_size_gb,
            additional_env=additional_env,
            image_name=image_name,
            send_status_email=send_status_email,
        )

    def _generate_job_params(self) -> list[str]:
        # Derive parameters

        positional_parameters, named_parameters = extract_parameters(self.udf.code)
        # Replace default code params with params passed to job
        named_parameters.update(self.udf.parameters)
        # Remove positional parameters that already have values in the job
        positional_parameters = [
            positional_parameter
            for positional_parameter in positional_parameters
            if positional_parameter not in named_parameters.keys()
        ]

        if self.type == "udf":
            if self.input:
                positional_parameters[0] = f"arg_list={repr(self.input)}"
        else:
            warnings.warn(
                FusedTypeWarning(f'Rendering of type "{self.type}" may be incomplete.')
            )

        # After replacing the inputs, look for any parameters with no values
        # that match the reserved parameter names. These should not be included
        # in instantiating the UDF because they are provided by the system.
        positional_parameters = [
            param
            for param in positional_parameters
            if param not in SYSTEM_PARAMETER_NAMES
        ]

        _params_fn_instance = positional_parameters + stringify_named_params(
            named_parameters
        )

        output_config = []
        if hasattr(self, "output"):
            output_str = stringify_output(self.output)
            if output_str is not None:
                output_config.append(f"output_table={output_str}")

        return _params_fn_instance + output_config

    def _generate_code(self, headerfile=False):
        # String: Job instantiation
        str_job_inst = f"job = {self.udf.entrypoint}({structure_params(self._generate_job_params())})"
        # String: Job execution
        str_job_exec = "job._run_local()"
        # String: UDF
        str_udf, header_cells = self.udf._generate_code(
            include_imports=False, headerfile=headerfile
        )
        # Structure cell
        src = f"""
{STR_IMPORTS}
{str_udf}
{str_job_inst}
{str_job_exec}
"""
        return src, header_cells

    def render(self, headerfile=False):
        _render(self, headerfile=headerfile)

    def export(
        self,
        path,
        how: Literal["local", "zip"] = "local",
        overwrite=False,
    ):
        _export(job=self, path=path, how=how, overwrite=overwrite)

    def set_output(
        self,
        table_or_url: str | None = None,
        *,
        table: str | None = None,
        url: str | None = None,
        inplace: bool = False,
        overwrite: bool | None = None,
    ) -> JobStepConfig:
        """Update output tables on this operation

        Args:
            table_or_url: Automatically set either `table` or `url` depending on whether this is a URL.

        Keyword Args:
            table: The name of the table to use for output. This table name must be unique. Defaults to None.
            url: If set, the URL to write the table to. Overrides `table` and `base_path`.
            inplace: If True, modify and return this object. If False, modify and return a copy. Defaults to False.
            overwrite: If True, overwrite the output dataset if it already exists. Defaults to None to not update.

        Returns:
            _description_
        """
        ret = _maybe_inplace(self, inplace)

        if not hasattr(ret, "output"):
            raise NotImplementedError("This job step does not have output")

        if table is None and url is None and table_or_url is not None:
            try:
                parsed = urlparse(table_or_url)
                if parsed.scheme:
                    url = table_or_url
            except (ValueError, TypeError, AttributeError):
                pass

            # Parsing as URL failed
            if url is None:
                table = table_or_url

        if url is not None and url is not Ellipsis:
            # If a bare URL is being set, it must be done on V2
            ret.output = DatasetOutputV2(
                url=url,
                save_index=ret.output.save_index,
                sample_strategy=ret.output.sample_strategy,
                overwrite=overwrite if overwrite is not None else ret.output.overwrite,
            )

        return ret


class PartitionJobStepConfig(JobStepConfig):
    """Base class for partitioner steps (should not be instantiated directly)"""

    input: list[StrictStr] | StrictStr
    """The path to the input file(s) for partitioning."""

    output: StrictStr | None = None
    """Base path of the main output directory"""
    output_metadata: StrictStr | None = None
    """Base path of the fused output directory"""

    partitioning_maximum_per_file: int | None = 2_500_000
    """Maximum value for `partitioning_method` to use per file. If `None`, defaults to _1/10th_ of the total value of `partitioning_method`. So if the value is `None` and `partitioning_method` is `"area"`, then each file will be have no more than 1/10th the total area of all geometries.
    """

    partitioning_maximum_per_chunk: int | None = 65000
    """Maximum value for `partitioning_method` to use per chunk. If `None`, defaults to _1/100th_ of the total value of `partitioning_method`. So if the value is `None` and `partitioning_method` is `"area"`, then each file will be have no more than 1/100th the total area of all geometries.
    """

    overwrite: bool = False
    """If True, overwrite the output directory if it already exists (by first removing
    all existing content of the directory). Defaults to False.
    """

    def _validate_for_run(
        self,
        *,
        ignore_no_udf: bool = False,
        ignore_no_output: bool = False,
        validate_imports: bool | None = None,
        validate_inputs: bool = True,
    ):
        super()._validate_for_run(
            ignore_no_udf=ignore_no_udf,
            ignore_no_output=ignore_no_output,
            validate_imports=validate_imports,
            validate_inputs=validate_inputs,
        )

        if validate_inputs:

            def _validate_input(input: str):
                if isinstance(input, str):
                    if (
                        input.startswith("file://")
                        or input.startswith("/")
                        or input.startswith("./")
                    ):
                        warnings.warn(
                            FusedPathWarning(
                                f"Input to partition job step ({input}) is a local file reference, which will be resolved on the backend and not on your local system. If you meant to partition a local file, use `fused.upload` first and then use the remote path."
                            )
                        )
                else:
                    warnings.warn(
                        FusedTypeWarning("Input to partition job step is not a string.")
                    )

            if isinstance(self.input, (list, tuple)):
                for input in self.input:
                    _validate_input(input)
            else:
                _validate_input(self.input)

    def _repr_html_(self) -> str:
        return fused_ingestion_repr(self)


class GDALOpenConfig(BaseModel):
    """A class to define options for how to open files with GDAL."""

    open_options: dict[str, str] = Field(default_factory=dict)
    """A dictionary of options passed in to GDAL for opening files."""

    layer: StrictStr | None = None
    """The layer of the input file to read from."""


class GeospatialPartitionJobStepConfig(PartitionJobStepConfig):
    type: Literal["partition_geospatial"] = "partition_geospatial"

    table_schema: Schema | None = None

    file_suffix: StrictStr | None = None
    load_columns: list[StrictStr] | None = None
    remove_cols: list[StrictStr] = Field(default_factory=list)
    explode_geometries: StrictBool = False

    drop_out_of_bounds: StrictBool | None = None
    """Whether to drop points that are outside of the WGS84 valid bounds."""

    lonlat_cols: tuple[str, str] | None = None
    """Names of longitude, latitude columns to construct point geometries from.

    This currently applies only to loading Parquet files.

    If the original files are in a format such as CSV, pass the names of the longitude
    and latitude columns in the GDALOpenConfig. If you pass those to GDALOpenConfig, do
    not also pass names to lonlat_columns here.
    """

    # Partitioning options
    partitioning_max_width_ratio: StrictFloat | StrictInt = 2
    partitioning_max_height_ratio: StrictFloat | StrictInt = 2

    partitioning_method: Literal["area", "length", "coords", "rows"]
    """The method used for deciding how to group geometries."""

    partitioning_force_utm: Literal["file", "chunk"] | None = None

    # TODO: switch back to median as default
    partitioning_split_method: Literal["mean", "median"]

    # Subdivide options
    subdivide_start: float | None = None
    """Geometries with greater area than this (in WGS84 degrees) will be subdivided.
    Start area should be greater than or equal to stop area.
    """

    subdivide_stop: float | None = None
    """This is the area that will stop continued subdivision of a geometry.
    Stop area should be less than or equal to start area. Additionally stop area cannot
    be zero, as that would cause infinite subdivision.
    """

    subdivide_method: Literal["area"] | None = "area"

    split_identical_centroids: StrictBool = True
    """
    Whether to split a partition that has identical centroids (such as if all geometries
    in the partition are the same) if there are more such rows than defined in
    "partitioning_maximum_per_file" and "partitioning_maximum_per_chunk".
    """

    target_num_chunks: StrictInt = 5000
    """The target for the number of chunks if partitioning_maximum_per_file is None."""

    partitioning_schema_input: StrictStr | None = None
    """Path to Parquet file with pre-defined partitioning schema"""

    gdal_config: GDALOpenConfig = Field(default_factory=GDALOpenConfig)
    """Options to pass to GDAL for opening files."""

    def run_batch(
        self,
        output_table: str | None = ...,
        instance_type: WHITELISTED_INSTANCE_TYPES | None = None,
        *,
        region: str | None = None,
        disk_size_gb: int | None = None,
        additional_env: list[str] | None = None,
        image_name: str | None = None,
        ignore_no_udf: bool = False,
        ignore_no_output: bool = False,
        validate_imports: bool | None = None,
        validate_inputs: bool = True,
        overwrite: bool | None = None,
        send_status_email: bool | None = None,
    ) -> RunResponse:
        """Execute this operation

        Args:
            output_table: The name of the table to write to. Defaults to None.
            instance_type: The AWS EC2 instance type to use for the job. Acceptable strings are "m5.large", "m5.xlarge", "m5.2xlarge", "m5.4xlarge", "m5.8xlarge", "m5.12xlarge", "m5.16xlarge", "r5.large", "r5.xlarge", "r5.2xlarge", "r5.4xlarge", "r5.8xlarge", "r5.12xlarge", or "r5.16xlarge". Defaults to None.
            region: The AWS region in which to run. Defaults to None.
            disk_size_gb: The disk size to specify for the job. Defaults to None.
            additional_env: Any additional environment variables to be passed into the job. Defaults to None.
            image_name: Custom image name to run. Defaults to None for default image.

            ignore_no_udf: Ignore validation errors about not specifying a UDF. Defaults to False.
            ignore_no_output: Ignore validation errors about not specifying output location. Defaults to False.
            send_status_email: Whether to send a status email to the user when the job is complete.
        """
        to_run = self.model_copy(deep=True)

        if output_table is Ellipsis:
            if self.output is None and self.output_metadata is None:
                raise ValueError("The Job requires `output_table` to be specified.")
            elif overwrite is not None:
                warnings.warn(
                    FusedIgnoredWarning("Overwrite on ingestion operation is ignored")
                )
        elif output_table is not None:
            to_run = to_run.set_output(
                output=output_table, overwrite=overwrite, inplace=False
            )
        else:
            ignore_no_output = True
        return to_run._run_remote(
            instance_type=instance_type,
            region=region,
            disk_size_gb=disk_size_gb,
            additional_env=additional_env,
            image_name=image_name,
            ignore_no_udf=ignore_no_udf,
            ignore_no_output=ignore_no_output,
            validate_imports=validate_imports,
            validate_inputs=validate_inputs,
            send_status_email=send_status_email,
        )

    run_remote = run_batch

    def set_output(
        self,
        output: str | None = None,
        output_metadata: str | None = None,
        inplace: bool = False,
        overwrite: bool | None = None,
    ):
        ret = _maybe_inplace(self, inplace)

        if overwrite is not None:
            warnings.warn(
                FusedIgnoredWarning("Overwrite on ingestion operation is ignored")
            )

        if output is not None and urlparse(output).scheme:
            ret.output = output
        else:
            raise ValueError(
                "Setting an invalid output table name for `output`. Set an output URL instead."
            )

        if output_metadata is not None and urlparse(output_metadata).scheme:
            ret.output_metadata = output_metadata
        else:
            raise ValueError(
                "Setting an invalid output table name for `output_metadata`. Set an output URL instead."
            )

        return ret


class NonGeospatialPartitionJobStepConfig(PartitionJobStepConfig):
    type: Literal["partition_nongeospatial"] = "partition_nongeospatial"

    partition_col: StrictStr | None = None


class UdfJobStepConfig(JobStepConfig):
    """A job step of running a UDF."""

    type: Literal["udf"] = "udf"
    udf: AnyBaseUdf

    input: list[Any] | None = None
    _validate_version: bool = True
    model_config = ConfigDict(populate_by_name=True)

    def set_input(
        self,
        input: list[Any] | None,
        inplace: bool = False,
    ) -> UdfJobStepConfig:
        """Set the input datasets on this operation

        Args:
            input: A list of JSON-serializable objects to pass as input to the UDF, or None to run once with no arguments.
            inplace: If True, modify and return this object. If False, modify and return a copy. Defaults to False.
        """
        ret = _maybe_inplace(self, inplace)
        ret.input = input

        return ret

    def set_udf(
        self,
        udf: AnyBaseUdf | dict | str,
        parameters: dict[str, Any] | None = None,
        replace_parameters: bool = False,
        inplace: bool = False,
    ) -> UdfJobStepConfig:
        """Set a user-defined function on this operation

        Args:
            udf: the representation of this UDF
            parameters: Parameters to set on the UDF. Defaults to None to not set parameters.
            replace_parameters: If True, unset any parameters not passed in parameters. Defaults to False.
            inplace: If True, modify and return this object. If False, modify and return a copy. Defaults to False.
        """
        ret = _maybe_inplace(self, inplace)
        ret.udf = RootAnyBaseUdf.model_validate(udf).root

        if parameters is not None:
            _assert_udf_has_parameters(ret.udf)
            ret.udf = ret.udf.set_parameters(
                parameters, replace_parameters=replace_parameters, inplace=False
            )

        return ret

    def _run_local(
        self,
        sample: Any | None = ...,
        validate_imports: bool | None = None,
        _return_response: bool = False,
        cache_max_age: int | None = None,
        **kwargs,
    ) -> UdfEvaluationResult:
        """
        Run a UDF locally on sample data.

        Args:
            sample: The sample input to pass to the UDF. Defaults to None.
            **kwargs: Additional keyword arguments to be passed to the UDF.

        Returns:
            The output of the user-defined function (UDF) applied to the input data.

        Raises:
            Any exceptions raised by the user-defined function (UDF) during its execution.
        """
        from fused._udf.execute_v2 import execute_against_sample

        default_sample = [] if not self.input else [self.input[0]]
        sample_list = default_sample if sample is Ellipsis else [sample]
        return execute_against_sample(
            self.udf,
            sample_list,
            validate_imports=validate_imports,
            _return_response=_return_response,
            cache_max_age=cache_max_age,
            **kwargs,
        )

    def _validate_for_run(
        self,
        *,
        ignore_no_udf: bool = False,
        ignore_no_output: bool = False,
        validate_imports: bool | None = None,
        validate_inputs: bool = True,
    ):
        if not ignore_no_udf and not self.udf.code:
            raise ValueError(
                "No UDF code is set. Set the `udf` attribute or pass `ignore_no_udf=True`."
            )

        # TODO: should _common_validate_before_run be used here?

    def _repr_html_(self) -> str:
        return fused_udf_step_repr(self)

    def get_sample(
        self,
        file_id: str | int = 0,
    ) -> Any:
        if isinstance(file_id, str):
            file_id = int(file_id)
        if self.input is not None and len(self.input) >= file_id:
            return self._run_local(self.input[file_id])
        else:
            raise ValueError("No input is set.")

    def _run_file(
        self,
        *,
        cache_max_age: int | None = None,
        _client_id: str | None = None,
        _format: str | None = None,
        _profile: bool = False,
        _run_id: str | None = None,
        _use_polling: bool = True,
        **kwargs,
    ) -> UdfEvaluationResult:
        # TODO: fix circular import error
        from fused._quick.udf import run as _run_realtime

        if _use_polling:
            from fused._quick.udf import run_polling

            res = run_polling(
                step_config=self,
                params=kwargs,
                client_id=_client_id,
                cache_max_age=cache_max_age,
                format=_format,
                _profile=_profile,
                run_id=_run_id,
            )
            return res

        res = _run_realtime(
            step_config=self,
            params=kwargs,
            client_id=_client_id,
            cache_max_age=cache_max_age,
            format=_format,
            _profile=_profile,
            run_id=_run_id,
        )
        return res

    async def _run_file_async(
        self,
        *,
        cache_max_age: int | None = None,
        _client_id: str | None = None,
        _format: str | None = None,
        _run_id: str | None = None,
        _use_polling: bool = True,
        **kwargs,
    ):
        # TODO: fix circular import error
        from fused._quick.udf import run_async as _run_realtime_async

        if _use_polling:
            from fused._quick.udf import run_polling_async

            res = await run_polling_async(
                step_config=self,
                params=kwargs,
                client_id=_client_id,
                cache_max_age=cache_max_age,
                format=_format,
                run_id=_run_id,
            )
            return res

        res = await _run_realtime_async(
            step_config=self,
            params=kwargs,
            client_id=_client_id,
            cache_max_age=cache_max_age,
            format=_format,
            run_id=_run_id,
        )
        return res

    # TODO: do we enforce pattern to have same args in overload, for typehints?
    def _run_tile(
        self,
        *,
        x: float = None,
        y: float = None,
        z: float = None,
        cache_max_age: int | None = None,
        _client_id: str | None = None,
        _format: str | None = None,
        _profile: bool = False,
        _run_id: str | None = None,
        _use_polling: bool = True,
        **kwargs,
    ) -> UdfEvaluationResult:
        from fused._quick.udf import run_tile as _run_realtime_tile

        if _use_polling:
            from fused._quick.udf import run_polling

            res = run_polling(
                step_config=self,
                params=kwargs,
                xyz=(x, y, z),
                client_id=_client_id,
                cache_max_age=cache_max_age,
                format=_format,
                _profile=_profile,
                run_id=_run_id,
            )
            return res

        res = _run_realtime_tile(
            step_config=self,
            x=x,
            y=y,
            z=z,
            client_id=_client_id,
            cache_max_age=cache_max_age,
            format=_format,
            _profile=_profile,
            run_id=_run_id,
            params=kwargs,
        )
        return res

    # TODO: do we enforce pattern to have same args in overload, for typehints?
    async def _run_tile_async(
        self,
        *,
        x: float = None,
        y: float = None,
        z: float = None,
        cache_max_age: int | None = None,
        _client_id: str | None = None,
        _format: str | None = None,
        _run_id: str | None = None,
        _use_polling: bool = True,
        **kwargs,
    ) -> UdfEvaluationResult:
        from fused._quick.udf import run_tile_async as _run_realtime_tile_async

        if _use_polling:
            from fused._quick.udf import run_polling_async

            res = await run_polling_async(
                step_config=self,
                params=kwargs,
                xyz=(x, y, z),
                client_id=_client_id,
                cache_max_age=cache_max_age,
                format=_format,
                run_id=_run_id,
            )
            return res

        res = await _run_realtime_tile_async(
            step_config=self,
            x=x,
            y=y,
            z=z,
            client_id=_client_id,
            cache_max_age=cache_max_age,
            format=_format,
            run_id=_run_id,
            params=kwargs,
        )
        return res

    def _run_batch(
        self,
        *,
        instance_type: WHITELISTED_INSTANCE_TYPES | None = None,
    ) -> UdfEvaluationResult:
        response = self.run_batch(instance_type=instance_type)
        print(f"Batch job started with ID: {response.job_id}")

        response = response.wait_for_job()
        if response.job_status in ("Failed", "Canceled"):
            raw_logs = response.get_logs()
            logs = "\n".join([m["message"].strip() for m in raw_logs])
            result = UdfEvaluationResult(
                data=None,
                time_taken_seconds=0.0,
                stdout=logs,
                has_exception=True,
                error_message=f"Batch job {response.job_status.lower()}",
                error_type="runtime_error",
            )
        elif response.job_status == "Done":
            results = response.get_results()
            if not results:
                # UDF returned None, TODO should handle this on the server side
                result = UdfEvaluationResult(
                    data=None,
                    time_taken_seconds=0.0,
                )
            else:
                result = results[0]
        else:
            raise RuntimeError(
                f"Unexpected job status: {response.job_status}. "
                "Expected 'Done', 'Failed', or 'Canceled'."
            )
        result.time_taken_seconds = response.get_exec_time()
        return result


AnyJobStepConfig = Annotated[
    Union[
        UdfJobStepConfig,
        GeospatialPartitionJobStepConfig,
        NonGeospatialPartitionJobStepConfig,
    ],
    Field(..., discriminator="type"),
]


class RootAnyJobStepConfig(RootModel[AnyJobStepConfig]):
    pass


class JobConfig(FusedBaseModel):
    name: StrictStr | None = None
    """The name of the job."""

    steps: list[AnyJobStepConfig]
    """The individual steps to run in sequence in the job."""

    metadata: UserMetadataType = None
    """User defined metadata. Any metadata values must be JSON serializable."""
    _validate_version: bool = True

    @property
    def udfs(self) -> UdfRegistry:
        udf_dict = {}
        for step in self.steps:
            udf_dict[step.udf.name] = step.udf
        return UdfRegistry(udf_dict)

    @field_validator("name")
    @classmethod
    def validate_job_name(cls, name: str | None) -> str | None:
        """
        Ensure that `name` is either None or matches /^[a-zA-Z0-9_-]+$/.
        If it’s invalid, raise ValueError.
        """
        if name is None:
            return None

        regex = r"[a-zA-Z][a-zA-Z0-9_.:\-]*"
        if not re.fullmatch(regex, name):
            raise ValueError(
                "Job name must be alphanumeric and may include dashes (-), underscores (_), period (.), and colon (:). "
                "Special characters and spaces are not allowed."
            )
        return name

    def _validate_for_run(
        self,
        *,
        ignore_no_udf: bool = False,
        ignore_no_output: bool = False,
        validate_imports: bool | None = None,
        validate_inputs: bool = True,
    ):
        for step in self.steps:
            step._validate_for_run(
                ignore_no_udf=ignore_no_udf,
                ignore_no_output=ignore_no_output,
                validate_imports=validate_imports,
                validate_inputs=validate_inputs,
            )

    def run_batch(
        self,
        instance_type: WHITELISTED_INSTANCE_TYPES | None = None,
        *,
        region: str | None = None,
        disk_size_gb: int | None = None,
        additional_env: list[str] | None = None,
        image_name: str | None = None,
        ignore_no_udf: bool = False,
        ignore_no_output: bool = False,
        validate_imports: bool | None = None,
        validate_inputs: bool = True,
        send_status_email: bool | None = None,
        **kwargs,
    ) -> RunResponse:
        """Execute an operation

        Keyword Args:
            region: The AWS region in which to run. Defaults to None.
            instance_type: The AWS EC2 instance type to use for the job. Acceptable strings are "m5.large", "m5.xlarge", "m5.2xlarge", "m5.4xlarge", "m5.8xlarge", "m5.12xlarge", "m5.16xlarge", "r5.large", "r5.xlarge", "r5.2xlarge", "r5.4xlarge", "r5.8xlarge", "r5.12xlarge", or "r5.16xlarge". Defaults to None.
            disk_size_gb: The disk size to specify for the job. Defaults to None.
            additional_env: Any additional environment variables to be passed into the job, each in the form KEY=value. Defaults to None.
            image_name: Custom image name to run. Defaults to None for default image.
            send_status_email: Whether to send a status email to the user when the job is complete.
        """
        assert self._api is not None

        # Operate on a copy because step parameters might change.
        ret = _maybe_inplace(self, inplace=False)

        # TODO: if the user uses start_job, this validation will never happen
        ret._validate_for_run(
            ignore_no_udf=ignore_no_udf,
            ignore_no_output=ignore_no_output,
            validate_imports=validate_imports,
            validate_inputs=validate_inputs,
        )

        for step in ret.steps:
            if hasattr(step, "udf"):
                filtered_kwargs = {
                    k: v for k, v in kwargs.items() if k in step.udf._parameter_list
                }

                step.udf.parameters.update(filtered_kwargs)

        return ret._api.start_job(
            config=self,
            instance_type=instance_type,
            region=region,
            disk_size_gb=disk_size_gb,
            additional_env=additional_env,
            image_name=image_name,
            send_status_email=send_status_email,
        )

    run_remote = run_batch

    def _to_job_step_config(self) -> JobStepConfig:
        if len(self.steps) > 1:
            warnings.warn(
                FusedDefaultWarning(
                    "input has more than one job step, taking the first"
                )
            )
        assert len(self.steps) > 0, "Input should have a job step defined"
        return self.steps[0]

    def _repr_html_(self) -> str:
        return fused_job_repr(self)

    def export(
        self,
        path,
        how: Literal["local", "zip"] = "local",
        overwrite=False,
    ):
        _export(job=self, path=path, how=how, overwrite=overwrite)

    def _generate_code(self, headerfile=False):
        udfs = {}
        header_cells_list = []
        jobs = []
        job_names = []
        for step in self.steps:
            str_udf, header_cells = step.udf._generate_code(
                include_imports=False, headerfile=headerfile
            )
            udfs[step.udf.name] = str_udf
            header_cells_list.extend(header_cells)
            # TODO: catch overlapping names
            # String: job instantiation
            job_names.append(f"job_{step.udf.name}")
            jobs.append(
                f"job_{step.udf.name} = {step.udf.name}({structure_params(step._generate_job_params())})"
            )

        # String: Job steps
        str_udfs = "\n\n".join(udfs.values())
        str_job = "\n".join(jobs)
        if len(job_names) == 1:
            str_multijob = f"job = fused.experimental.job({job_names[0]})"
        else:
            str_multijob = f"job = fused.experimental.job([{', '.join(job_names)}])"
        # String: Job execution
        str_job_exec = "job.run_local()"
        # Structure cell
        src = f"""
{STR_IMPORTS}
{str_udfs}\n
{str_job}
{str_multijob}
{str_job_exec}
"""
        return src, set(header_cells_list)

    def run_local(
        self,
        validate_imports: bool | None = None,
        *args,
        **kwargs,
    ) -> MultiUdfEvaluationResult:
        # For each step, run the JobStepConfig with run_local
        runs = []
        for step in self.steps:
            filtered_kwargs = {
                k: v for k, v in kwargs.items() if k in step.udf._parameter_list
            }
            _step = step._run_local(
                validate_imports=validate_imports, *args, **filtered_kwargs
            )
            runs.append(_step)
        return MultiUdfEvaluationResult(udf_results=[run for run in runs])

    def render(self, headerfile=False):
        _render(self, headerfile=headerfile)

    def get_sample(self):
        raise NotImplementedError(
            "Cannot get sample of a job. Please call get_sample on a job step. For example: `job.steps[0].get_sample()`"
        )


def _validate_headers_for_remote_exec(udf: AnyBaseUdf) -> bool:
    """If unresolved headers reference local files, returns False."""

    for header in udf.headers:
        # If header is a string that is not a remote path.
        if isinstance(header, str) and not is_url(header):
            warnings.warn(FusedWarning(f"Header {header} cannot resolve remotely."))
            return False

    return True


def _export(job, path, how, overwrite=False):
    # Validate `how`
    if how not in ("local", "zip"):
        raise ValueError("`how` must be one of 'local', or 'zip'")
    # Handle single step jobs
    if isinstance(job, JobStepConfig):
        job = JobConfig(steps=[job])

    files = {
        "meta.json": generate_meta_json(job),
        # "multijob.py": obj._generate_code(headerfile=True),
        "README.md": generate_readme(job),
    }

    # Files: UDFs & Headers
    for step in job.steps:
        for header in step.udf.headers:
            # Don't raise warning if header names overlap, since they're shared.
            header_filename = header.module_name + ".py"
            if header_filename not in files:
                src = header.source_code
                files[header_filename] = src
        # Raise warning if UDF names overlap.
        udf_filename = f"{step.udf.name}.py"
        if udf_filename not in files:
            files[udf_filename] = STR_IMPORTS + step.udf.code
        else:
            warnings.warn(
                FusedDefaultWarning(f"Duplicate UDF name {step.udf.name}. Skipping.")
            )

    # Export
    create_directory_and_zip(path=path, how=how, files=files, overwrite=overwrite)


def _render(job, headerfile=False):
    from IPython import get_ipython
    from IPython.core.inputsplitter import IPythonInputSplitter

    def create_new_cell(contents):
        """Similar to ipython.set_next_input but allows multiple cells to be created at once."""
        from IPython.core.getipython import get_ipython

        shell = get_ipython()

        payload = dict(
            source="set_next_input",
            text=contents,
            replace=False,
        )
        shell.payload_manager.write_payload(payload, single=False)

    # Get the current IPython instance.
    ipython = get_ipython()
    if ipython is None:
        raise RuntimeError("This function can only be used in a Jupyter Notebook.")

    # Create an instance of IPythonInputSplitter.
    splitter = IPythonInputSplitter()

    # Generate code string and split into lines.
    src_udf, header_cells = job._generate_code(headerfile=headerfile)
    lines = src_udf.strip().split("\n")

    # Set the content of the subsequent cell with.
    create_new_cell(splitter.transform_cell("\n".join(lines)))

    # Headers
    if headerfile:
        for header in header_cells:
            create_new_cell(header)


class DatasetInputBase(BaseModel):
    type: Literal[None, "v2"]

    def first_partition(self) -> tuple[str, int]:
        return "0", 0



import ast
from typing import Optional

from ..internal import DatasetOutputV2


def stringify_named_params(params):
    return [k + "=" + repr(v) for k, v in params.items()]


def structure_params(params, separator=", "):
    return separator.join([param for param in params])


def stringify_headers(headers):
    if not headers:
        return "[]"
    else:
        for header in headers:
            if hasattr(header, "source_file"):
                delattr(header, "source_file")
        return str(headers)


def stringify_output(output) -> Optional[str]:
    if isinstance(output, DatasetOutputV2):
        return repr(output.url) if output.url is not None else None

    return repr(output)


def extract_parameters(src):
    # Parse the input string into an AST (Abstract Syntax Tree)
    parsed_ast = ast.parse(src)

    all_parameters = []
    named_parameters = {}

    # Find all function definitions in the AST
    function_defs = [
        node
        for node in ast.walk(parsed_ast)
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
    ]
    # Assume the first function is the target function
    target_function = function_defs[0]

    for arg in target_function.args.args:
        param_name = arg.arg
        # named_parameters[param_name] = None  # Initialize with None
        all_parameters.append(param_name)

    for keyword in target_function.args.kwonlyargs:
        param_name = keyword.arg
        # named_parameters[param_name] = None  # Initialize with None

    for param, default in zip(
        target_function.args.args[
            len(target_function.args.args) - len(target_function.args.defaults) :
        ],
        target_function.args.defaults,
    ):
        # Extract the default value if it's a string, number, or None
        # TODO: Handle more types
        if isinstance(
            default,
            (
                ast.Constant,
                ast.Tuple,
                ast.List,
                ast.Dict,
                ast.Set,
            ),
        ):
            named_parameters[param.arg] = ast.literal_eval(default)

    positional_parameters = [
        param for param in all_parameters if param not in named_parameters
    ]

    return positional_parameters, named_parameters



# TODO: This file is no longer the most recent -- use fused.core.run_* instead
# This file is only for running non-saved (code included) UDFs
from __future__ import annotations

import json
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

from loguru import logger

from fused._options import default_serialization_format
from fused._options import options as OPTIONS
from fused._request import session_with_retries
from fused.api.api import FusedAPI, resolve_udf_server_url
from fused.core._realtime_ops import (
    _process_response,
    _process_response_async,
    serialize_realtime_params,
)
from fused.models import AnyJobStepConfig
from fused.models.udf._eval_result import UdfEvaluationResult

from ..core._impl._realtime_ops_impl import get_recursion_factor
from ..core._realtime_ops import _realtime_raise_for_status_async

if TYPE_CHECKING:
    import pandas as pd


@dataclass
class RequestInfo:
    udf_server_url: str
    url: str
    body: dict
    headers: dict
    params: dict


def _prepare_request(
    step_config: AnyJobStepConfig | None,
    params: dict[str, str] | None,
    xyz: tuple[float, float, float] | None,
    client_id: str | None,
    cache_max_age: int | None,
    format: str | None,
    _profile: bool,
    run_id: str | None,
):
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    # Note: Custom UDF uses the json POST attribute.
    if xyz is None:
        url = f"{udf_server_url}/api/v1/run/udf"
    else:
        x, y, z = xyz
        url = f"{udf_server_url}/api/v1/run/udf/tiles/{z}/{x}/{y}"

    # Payload
    # This is the body for when step_config_with_params.type == "udf".
    body = {
        "step_config": step_config_with_params.model_dump_json(),
        "format": format or default_serialization_format(),
        "profile": _profile,
    }

    # Headers
    recursion_factor = get_recursion_factor()
    headers = api._generate_headers(
        {
            "Content-Type": "application/json",
            **(OPTIONS.default_run_headers or {}),
        }
    )
    headers["Fused-Recursion"] = f"{recursion_factor}"
    if run_id is not None:
        headers["Fused-Run-Id"] = run_id

    # Params
    req_params = {
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
    }

    return step_config_with_params, RequestInfo(
        udf_server_url=udf_server_url,
        url=url,
        body=body,
        headers=headers,
        params=req_params,
    )


def run_tile(
    x: float,
    y: float,
    z: float,
    step_config: AnyJobStepConfig | None = None,
    params: dict[str, str] | None = None,
    *,
    client_id: str | None = None,
    cache_max_age: int | None = None,
    format: str | None = None,
    _profile: bool = False,
    run_id: str | None = None,
) -> UdfEvaluationResult:
    return run(
        step_config=step_config,
        params=params,
        xyz=(x, y, z),
        client_id=client_id,
        cache_max_age=cache_max_age,
        format=format,
        _profile=_profile,
        run_id=run_id,
    )


def run(
    step_config: AnyJobStepConfig | None = None,
    params: dict[str, str] | None = None,
    xyz: tuple[float, float, float] | None = None,
    *,
    client_id: str | None = None,
    cache_max_age: int | None = None,
    format: str | None = None,
    _profile: bool = False,
    run_id: str | None = None,
) -> pd.DataFrame:
    """Run a UDF.

    Args:
        step_config: AnyJobStepConfig.
        params: Additional parameters to pass to the UDF. Must be JSON serializable.
    """
    time_start = time.perf_counter()

    # Prepare request body/headers/parameters
    step_config_with_params, request = _prepare_request(
        step_config=step_config,
        params=params,
        xyz=xyz,
        client_id=client_id,
        cache_max_age=cache_max_age,
        format=format,
        _profile=_profile,
        run_id=run_id,
    )

    # Make request
    start = time.time()

    with session_with_retries() as session:
        logger.debug(f"{run_id=} | Starting request")
        r = session.request(
            method="POST",
            url=request.url,
            params=request.params,
            json=request.body,
            headers=request.headers,
            timeout=OPTIONS.run_timeout,
        )

    end = time.time()
    logger.info(f"{run_id=} | Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return _process_response(
        r,
        step_config=step_config_with_params,
        time_taken_seconds=time_taken_seconds,
        profile=_profile,
    )


def run_polling(
    step_config: AnyJobStepConfig | None = None,
    params: dict[str, str] | None = None,
    *,
    xyz: tuple[float, float, float] | None = None,
    client_id: str | None = None,
    cache_max_age: int | None = None,
    format: str | None = None,
    _profile: bool = False,
    run_id: str | None = None,
) -> UdfEvaluationResult:
    time_start = time.perf_counter()

    # Prepare request body/headers/parameters
    step_config_with_params, request = _prepare_request(
        step_config=step_config,
        params=params,
        xyz=xyz,
        client_id=client_id,
        cache_max_age=cache_max_age,
        format=format,
        _profile=_profile,
        run_id=run_id,
    )

    if xyz is None:
        url = f"{request.udf_server_url}/api/v1/run/udf/start"
    else:
        x, y, z = xyz
        url = f"{request.udf_server_url}/api/v1/run/udf/start/tiles/{z}/{x}/{y}"

    # Make request
    start = time.time()

    # use request_timeout instead of run_timeout since the request does not
    # connect for the entirety of the UDF run
    timeout = OPTIONS.request_timeout

    with session_with_retries() as session:
        logger.debug(f"{run_id=} | Starting request")
        r = session.request(
            method="POST",
            url=url,
            params=request.params,
            json=request.body,
            headers=request.headers,
            timeout=timeout,
        )

        if r.status_code == 202:
            logger.debug(f"{run_id=} | Start polling")

            n_retries = 0
            while (time.time() - start) < OPTIONS.run_timeout:
                r = session.get(
                    f"{request.udf_server_url}/api/v1/run/udf/get/{run_id}",
                    headers=request.headers,
                    timeout=timeout,
                )
                if r.status_code != 202:
                    break
                if n_retries > 1:
                    time.sleep(1)
                n_retries += 1

    if r.status_code == 202:
        # TODO: Add a better error message
        raise Exception(f"UDF run {run_id} timed out")

    end = time.time()
    logger.info(f"{run_id=} | Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return _process_response(
        r,
        step_config=step_config_with_params,
        time_taken_seconds=time_taken_seconds,
        profile=_profile,
    )


async def run_tile_async(
    x: float,
    y: float,
    z: float,
    step_config: AnyJobStepConfig | None = None,
    params: dict[str, str] | None = None,
    *,
    client_id: str | None = None,
    cache_max_age: int | None = None,
    format: str | None = None,
    run_id: str | None = None,
) -> UdfEvaluationResult:
    return await run_async(
        step_config=step_config,
        params=params,
        xyz=(x, y, z),
        client_id=client_id,
        cache_max_age=cache_max_age,
        format=format,
        run_id=run_id,
    )


async def run_async(
    step_config: AnyJobStepConfig | None = None,
    params: dict[str, str] | None = None,
    *,
    xyz: tuple[float, float, float] | None = None,
    client_id: str | None = None,
    cache_max_age: int | None = None,
    format: str | None = None,
    run_id: str | None = None,
) -> UdfEvaluationResult:
    """Run a UDF over a DataFrame.

    Args:
        step_config: AnyJobStepConfig.
        params: Additional parameters to pass to the UDF. Must be JSON serializable.

    """
    time_start = time.perf_counter()

    # Prepare request body/headers/parameters
    step_config_with_params, request = _prepare_request(
        step_config=step_config,
        params=params,
        xyz=xyz,
        client_id=client_id,
        cache_max_age=cache_max_age,
        format=format,
        _profile=False,
        run_id=run_id,
    )

    # Make request
    start = time.time()

    if OPTIONS.pyodide_async_requests:
        import pyodide.http
        import yarl

        url_with_params = yarl.URL(request.url, encoded=True).with_query(request.params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            method="POST",
            headers=request.headers,
            body=json.dumps(request.body),
            # TODO: timeout
        )
        end = time.time()
        logger.info(f"Time in request: {end - start}")

        await _realtime_raise_for_status_async(r)
        time_end = time.perf_counter()
        time_taken_seconds = time_end - time_start

        return await _process_response_async(
            r,
            step_config=step_config_with_params,
            time_taken_seconds=time_taken_seconds,
        )
    else:
        from fused.core._realtime_ops import _get_shared_session

        # Use shared session for connection pooling and reuse
        # TODO: Retry mechanism
        session = await _get_shared_session()
        async with session.post(
            url=request.url,
            params=request.params,
            json=request.body,
            headers=request.headers,
            # retry=default_retry_strategy,
            # TODO: timeout
        ) as r:
            end = time.time()
            logger.info(f"Time in request: {end - start}")

            await _realtime_raise_for_status_async(r)
            time_end = time.perf_counter()
            time_taken_seconds = time_end - time_start

            return await _process_response_async(
                r,
                step_config=step_config_with_params,
                time_taken_seconds=time_taken_seconds,
            )


async def run_polling_async(
    step_config: AnyJobStepConfig | None = None,
    params: dict[str, str] | None = None,
    *,
    xyz: tuple[float, float, float] | None = None,
    client_id: str | None = None,
    cache_max_age: int | None = None,
    format: str | None = None,
    _profile: bool = False,
    run_id: str | None = None,
) -> UdfEvaluationResult:
    time_start = time.perf_counter()

    # Prepare request body/headers/parameters
    step_config_with_params, request = _prepare_request(
        step_config=step_config,
        params=params,
        xyz=xyz,
        client_id=client_id,
        cache_max_age=cache_max_age,
        format=format,
        _profile=_profile,
        run_id=run_id,
    )
    if xyz is None:
        url = f"{request.udf_server_url}/api/v1/run/udf/start"
    else:
        x, y, z = xyz
        url = f"{request.udf_server_url}/api/v1/run/udf/start/tiles/{z}/{x}/{y}"

    # Make request
    start = time.time()

    if OPTIONS.pyodide_async_requests:
        import pyodide.http
        import yarl

        logger.debug(f"{run_id=} | Starting request")
        url_with_params = yarl.URL(url, encoded=True).with_query(request.params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            method="POST",
            headers=request.headers,
            body=json.dumps(request.body),
            # TODO: timeout
        )
        if r.status == 202:
            logger.debug(f"{run_id=} | Start polling")

            url = yarl.URL(
                f"{request.udf_server_url}/api/v1/run/udf/get/{run_id}", encoded=True
            )

            n_retries = 0
            while (time.time() - start) < OPTIONS.run_timeout:
                r = await pyodide.http.pyfetch(
                    str(url),
                    method="GET",
                    headers=request.headers,
                )
                if r.status != 202:
                    break
                if n_retries > 1:
                    time.sleep(1)
                n_retries += 1

        result_content = await r.bytes()

    else:
        from fused.core._realtime_ops import _get_shared_session

        logger.debug(f"{run_id=} | Starting request")
        result_content: bytes | None = None
        session = await _get_shared_session()
        async with session.post(
            url=url,
            params=request.params,
            json=request.body,
            headers=request.headers,
        ) as r:
            if r.status == 202:
                logger.debug(f"{run_id=} | Start polling")

                n_retries = 0
                while (time.time() - start) < OPTIONS.run_timeout:
                    async with session.get(
                        url=f"{request.udf_server_url}/api/v1/run/udf/get/{run_id}",
                        headers=request.headers,
                    ) as r:
                        if r.status != 202:
                            result_content = await r.read()
                            break
                        if n_retries > 1:
                            time.sleep(1)
                        n_retries += 1
            else:
                result_content = await r.read()

    if r.status == 202:
        # TODO: Add a better error message
        raise Exception(f"UDF run {run_id} timed out")

    end = time.time()
    logger.info(f"{run_id=} | Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return await _process_response_async(
        r,
        step_config=step_config_with_params,
        time_taken_seconds=time_taken_seconds,
        result_content=result_content,
    )



from __future__ import annotations

from fused._udf.context import get_global_context


def context_get_user_email() -> str:
    context = get_global_context()
    assert context is not None, "context is unexpectedly None"
    assert hasattr(context, "user_email"), (
        "could not detect user ID from context, please specify the UDF as 'user@example.com', 'udf_name'."
    )
    return context.user_email


async def context_get_user_email_async() -> str:
    context = get_global_context()
    assert context is not None, "context is unexpectedly None"
    assert hasattr(context, "user_email_async"), (
        "could not detect user ID from context, please specify the UDF as 'user@example.com/udf_name'."
    )
    return await context.user_email_async()


def context_get_auth_header(*, missing_ok: bool = False) -> dict[str, str]:
    context = get_global_context()
    if not context:
        raise ValueError("missing global context")

    return context.auth_header(missing_ok=missing_ok)


def context_get_auth_scheme_and_token() -> tuple[str, str] | None:
    context = get_global_context()
    if not context:
        return None

    if context.auth_scheme and context.auth_token:
        return (context.auth_scheme, context.auth_token)
    return None


def context_in_realtime():
    context = get_global_context()
    return context.in_realtime


def context_in_batch():
    context = get_global_context()
    return context.in_batch
